"""
all the units for the values in the description files

this module is for import *

todo:

actually test the mixed timecode type code, which probably does work

gt, lt for timestamp

either make hash work, or declare it not-allowed

allow expressions for input, so i can tweak an xml file by putting
'+5fr' after a time, for example



(false,true)[not not expr]

expr ? true : false



"""
from __future__ import division
import time as timemodule
import re,sys,string
from cuisine.external.rational import rational

try:
    from xml.utils import iso8601
except ImportError:
    raise ImportError("No 'utils' module within 'xml' module. Perhaps you don't have PyXML installed (pyxml.sf.net)")


from timecode import Timecode

import logging
log=logging.getLogger("desctypes")


class Timestamp:
    """we're rounding to the second, which is what I gathered from the
    spec. the iso8601 module would actually include 1/100s of a
    second.

    the timezone selection in __str__ is untested. it works for me on
    october 1, 2002, at least. at worst, you'll get files encoded with
    a tz that's not your own, which is not too horrible.

    """
    def __init__(self,time="now",**kwargs):
        if time=="now":
            self._time=int(timemodule.time())
        else:
            try:
                self._time=int(iso8601.parse(time))
            except (ValueError,TypeError):

                raise ValueError
                # we stop here since decodeunit wants to try a bunch of __inits__
                # to find one that works
#                # consider the value to be seconds-from-epoch
#                self._time=float(time)
    def __float__(self):
        return self._time
    def __int__(self):
        return int(self._time)
    def __str__(self):
        tz=timemodule.timezone-timemodule.daylight*60*60
        return iso8601.tostring(self._time,tz)
    def __repr__(self):
        return "Timestamp('%s')" % self.__str__()

class Unum2(object):
    """base class for numbers with units. Subclass Unum2 to make a
    particular type of units (pun intended). you should subclass Unum2
    to make a certain 'dimension', such as length or mass.
    
    Unum2 is in the sprit of Unum (written by someone else), which
    seems not to be available anymore. I did not try to keep
    compatibility with Unum in any way (although a lot of operations
    do work the same way). In particular, Unum supported multiplying
    and dividing Unums to make Unums with new units. Unum2 does not
    (as of now) support that.

    things to override in subclasses:

      override the 'units' structure to describe the units that your
      type supports.  use multipliers to relate the units that are
      easily converted. use None for units that need special
      conversions, and override the convert() method to handle those.

      override convert() to handle fancier conversions, such as those
      that require fancier math (such as F<->C temperatures) or extra
      context (such as pixels<->centimeters)

      override parse() (and maybe __str__) if any of your units can
      appear in special forms that don't fit the regexp in Unum2.parse
      (which covers 5, 5px, 1/5px, 5.1px)

    throughout the class, None means 'unknown' or 'unset' or 'N/A'.

    internally, the _val and _unit attributes describe the entire value.

    Unum2 is not (directly) suitable for units that can't be added
    together, such as dates.

    """

    """these are the units this class can appear as. the dictionary
    keys are the full names of the units; the values are tuples of
    multipliers (or None) and tuples of the short forms of the unit
    name. the first short form is preferred. a multiplier of None
    indicates the conversion is a special case handled in the
    convert() method. special cases may be fancier math (such as F<->C
    temperatures) or conversions that require context (such as
    pixels<->centimeters)"""
    __slots__=('units','_val','_unit','inited','initkwargs')

    units={# no units in the base class
           #'meters':(1.0,('m',)),
           #'centimeters':(.01,('cm',)),
           #'feet':(.1,('ft','f')),
           }

    # this is a cache of all the values that Unum2 creates,
    # which let it init new values quickly (if they've been seen before).
    # this took about 3s off a 21s playback, which is nice. (there were
    # no cache dumps during that time). the cache dump algorithm is very
    # stupid currently. look in __init__ to see what it is.
    # also, this cache has a side effect of keeping refs around for
    # a while, which might be sucky.
    cache_values = {} # args : Unum2

    def __new__(cls,*args,**kwargs):

        # turned this off since i was getting Unums with unset _vals!
        if 0 and args in Unum2.cache_values and not kwargs:
            ret=Unum2.cache_values[args]
            return ret
    
        new=super(Unum2,cls).__new__(cls,*args,**kwargs)
        Unum2.cache_values[args]=new
        return new

    def __init__(self,*args,**kwargs):
        """takes the given Unum2 or string and creates a new Unum2. If the string
        cannot be parsed, we raise a ValueError. That way, you can take an unknown
        value and attempt to make Unum2 (subclass) objects out of it, starting from
        the strictest syntax to the loosest."""

        if hasattr(self,'inited'):
            return
        
        self.inited=1

        self.initkwargs=kwargs # save these for other methods to use (see Duration/timecode for an example)

        #
        # this will fill in the val/unit from previous Unums we've made
        ###
##         if 0 and args in Unum2.cache_values and not kwargs:
##             hit=Unum2.cache_values[args]
##             self._val,self._unit=hit._val,hit._unit
##             return
##         if len(Unum2.cache_values.keys())>2000:
##             log.info("dumping Unum2 cache")
##             Unum2.cache_values.clear()
        ###

        if len(args)==0:
            # no args makes a zero
            (self._val,self._unit)=0,None
        elif len(args)==1:
            arg=args[0]
            if arg is None:
                raise ValueError("can't make a Unum2 from a None")
            if isinstance(arg,(str,unicode)):
                # one arg is a string like "5m"
                (self._val,self._unit)=self.parse(arg)
            elif isinstance(arg,Unum2):
                # or maybe a Unum2 to copy
                self._val = arg._val
                self._unit = arg._unit
            else:
                # some bad arg
                log.error("can't decode '%r'" % arg)
                raise ValueError("can't decode single-arg '%r'" % arg)
        elif len(args)==2:
            # two args is a value,unit pair. use any name style for the unit.
            v=args[0]
            if isinstance(v,(str,unicode)): 
                v=float(v) # sucky - this should eval for the proper numeric type
            (self._val,self._unit)=v,self.findunit(args[1])[0]
        else:
            raise ValueError("wrong number of arguments to constructor")

        Unum2.cache_values[args]=self

    #
    # methods to access the units structure
    #
    def findunit(cls,unitname):
        """returns the canonicalname,shortname for the
        given name (which can be either length).  case insensitive
        search."""
        if unitname is None:
            return None,None
        # it's faster if you use the canonicalname exactly
        if unitname in cls.units:
            return unitname,cls.units[unitname][1][0]
        # but you can any short name if you want to
        for k,v in cls.units.items():
            if unitname in v[1]:
                return k,v[1][0]
        if 0:
            # this is the case insens check, which is even slower
            for k,v in cls.units.items():
                if k.lower()==unitname.lower() or unitname.lower() in map(string.lower,v[1]):
                    return k,v[1][0] # k, first short name

        raise ValueError("unit '%s' unknown to type %s" % (unitname,cls.__name__))
    findunit=classmethod(findunit)
    def getmultiplier(cls,canonname):
        """returns the multiplier field from the units dictionary for the given
        full name, which is also a dict key"""
        return cls.units[canonname][0]
    getmultiplier=classmethod(getmultiplier)
    
    def parse(self,input):
        """parses string input into value and unit tuple, where value
        is a float and unit is a key in the units dictionary. raises
        ValueError (probably) if the input can't be parsed.

        if you want to handle additional formats, you'd probably want
        to try your formats and then fall back to this one."""

        try:
            (number,unit)=re.match(r'^([0-9\./-]+)([a-zA-Z]+)$',input).groups()
        except AttributeError:
            raise ValueError("input string didn't match regexp")

        # convert to int, float, or rational as appropriate
        i=number.find('/')
        if i!=-1:
            number=rational(num=int(number[:i]),den=int(number[i+1:]))
        else:
            i=number.find('.')
            if i==-1:
                number=int(number) # it might help precision sometimes to keep ints
            else:
                number=float(number)

        # make sure it's a unit we know, and change to the full name
        if unit!="":
            canon,short=self.findunit(unit)
            unit=canon
        else:
            unit=None # "unset", probably okay for zero

        return (number,unit)

    def convert(self,tounitname): # should be a canonical name
        """returns the value of this object in the new unit, raising
        TypeError if the unit is unknown or dissimilar.

        if you have special units that require custom conversion code,
        try this convert function first, then do your code. (or the
        other way around; it probably doesn't matter).

        NOT IMPLEMENTED (AND TBD):
          if this convert raises a TypeError, the policy is to try the
          conversion the other way. now, where desired, a finer value
          can be compared to a coarser value: the coarser value gives
          TypeError on conversion to the finer value, and the convert
          is reversed. this is getting to be like a coerce() call, and
          should be researched further.
        """

        # shortcut for conversion to the same unit
        if self._unit==self.findunit(tounitname)[0]:
            return self._val

        # attempt to use multiplication to convert
        m1=self.getmultiplier(self._unit)
        m2=self.getmultiplier(tounitname)

        return self._val*m1/m2
    
    def as(self,tounitname):
        """returns a copy of this object in the new unit"""
        assert isinstance(tounitname,str)
        canon,short=self.findunit(tounitname)
        new=self.__class__(self.convert(canon),canon)
        return new
    def getas(self,tounitname):
        """returns the value part of this object in the given units"""
        assert isinstance(tounitname,str)
        canon,short=self.findunit(tounitname)
        return self.convert(canon)
    def __str__(self):
        canon,short=self.findunit(self._unit)
        return "%s%s"% (self._val,short)
    def __repr__(self):
        return "%s(%r,%r)" % (self.__class__.__name__,self._val,self._unit)



    #
    # the math overriding functions - see after the class decl for the assignments 
    #
    def sameunitmath(self,other,fnname):
        "perform a function like '__add__'"
        if not isinstance(other,Unum2):
            raise ValueError("can't %s a %s and a %s" % (fnname,type(self),type(other)))
        fn=getattr(self._val,fnname)
        otherval=other.convert(self._unit) # (might throw exception)
        newval=fn(otherval)
        if newval is NotImplemented:
            newval=getattr(otherval,"__r"+fnname[2:])(self._val)
        if newval is NotImplemented:
            raise ValueError("math function not implemented")
        return self.__class__(newval,self._unit)

    def cmpmath(self,other,fnname):
        """perform a function like __cmp__ (returns a simple value
        with no units).  there is a coerce required in here, since
        self._val and the other value (after being normalized to the
        same units) need to be the same type for __cmp__. similar
        coercion may be needed in the other math functions here, i
        havent tested them yet.

        testing if a Unum2 __eq__ None returns 0. cmp with None is not defined.
        """
        
        if not isinstance(other,Unum2):
            if other is None and fnname=='__eq__': # can't cmp with None, but we can eq it.
                return 0
            raise ValueError("can't %s a %s and a %s" % (fnname,type(self),type(other)))
        otherval=other.convert(self._unit)
        x,y=coerce(self._val,otherval)
        fn=getattr(x,fnname,None)
        if fn is None:
            return NotImplemented # ints dont have __eq__, for example
        return fn(y)

    def scalemath(self,other,fnname):
        "perform a function like '__mul__' with a unitless number"
        fn=getattr(self._val,fnname)
        newval=fn(other)
        return self.__class__(newval,self._unit)

    def divmath(self,other,fnname):
        "handle division"
        if isinstance(other,Unum2):
            # Unum2/Unum2 becomes unitless
            otherval=other.convert(self._unit) # (but first, the units have to be coerceable)
            return getattr(self._val,fnname)(otherval)
        else:
            # Unum2/float, for example
            return self.scalemath(other,fnname)

    def unarymath(self,fnname):
        "perform a function like '__neg__' on our value"
        fn=getattr(self._val,fnname)
        return self.__class__(fn(),self._unit)
    
#
# class Unum2's not over yet!
# Now we assign the 3 math handling methods to all the
# math functions we're going to support.
#
# same-unit math
for attr in ["__%s__"%x for x in 'add sub'.split()]:
    setattr(Unum2,attr,lambda self,other,fnname=attr: self.sameunitmath(other,fnname))
# scale math
for attr in '__mul__','__rmul__':
    setattr(Unum2,attr,lambda self,other,fnname=attr: self.scalemath(other,fnname))
# dividing math
fns='floordiv div truediv'.split()
for attr in ["__%s__"%x for x in fns]+["__r%s__"%x for x in fns]:
    setattr(Unum2,attr,lambda self,other,fnname=attr: self.divmath(other,fnname))
# unary math
for attr in '__neg__','__pos__','__abs__','__invert__':
    setattr(Unum2,attr,lambda self,fnname=attr: self.unarymath(fnname))
# cmp
for attr in ('__cmp__','__eq__'):
    setattr(Unum2,attr,lambda self,other,fnname=attr: self.cmpmath(other,fnname))
# in-place math falls back on the other functions, which is
# fine, since we're staying immutable
# Please see http://python.org/doc/ref/coercion-rules.html for more.


##############################################################
## Unum2 subclasses for various dimensions

class Length(Unum2):
    """
    a length, not a duration or a filesize. length can be in pixels,
    cm, etc.  conversion may require context. for example, 50px by
    50px might be 40u by 30u (where u is some square screen unit) if
    the current aspect is 4/3.
    
    """
    units={'meters':(1.0,('m',)),
           'centimeters':(.01,('cm',)),
           'feet':(.3048,('ft','f',"'")),
           'inches':(.0254,('in','"')),
           'pixels':(None,('px',)),
           }

class Duration(Unum2):
    """
    time duration or offset. this has context-- 15fr might be .5s in
    one context but 3s in another.  different origins are not handled
    here: if you want 0fr to have timecode 01:04:02:00, you need to
    remember that yourself.

    timecode stores as a signed integer frame count internally, and str
    defaults to displaying it as a [-]00:00:00:00 string. frames can be
    fractional, timecode can't.  converting to timecode performs a
    round-off. *comparing frames and timecode is done in integer
    space-- 5.5 frames==00:00:00:05*. this should make everything work
    well enough for now, while allowing for frac-frame edits in the
    future.

    don't use Timecode(...), use Duration(...) so math gets supported right.
    you can still pass in timecodetype='24' or whatever.
    
    """
    __slots__=()
    units={'seconds':(1.0,('s','sec','secs')),
           'minutes':(60,('m','min','mins')),
           'hours':(60*60,('h','hr','hrs')),
           'frames':(None,('fr',)),
           'timecode':(None,('',)),
           }
    def parse(self,input):
        try:
            if not re.match("^-?\d+:\d\d:\d\d:\d\d$",input):
                raise ValueError # if it's not a timeline value, abort this special case and do a normal parse
            # timecodes are stored as integer numbers of frames
            neg=(input[0]=="-")
            ret= (((-1)**neg)*int(Timecode(input[neg:],**self.initkwargs)),'timecode')
            return ret
        except ValueError:
            return super(Duration,self).parse(input)
    def __str__(self):
        if self._unit=="timecode":
            # format the frame count as a string
            if self._val<0:
                return "-"+str(Timecode(int(-self._val),**self.initkwargs))
            else:
                return str(Timecode(int(self._val),**self.initkwargs))
        else:
            return super(Duration,self).__str__()
    def convert(self,tounitname):
        # timecode is stored with frames as the value, so there's
        # nothing to convert between timecode and frames (if the
        # frames are at the same framerate)
        if ((self._unit=="timecode" and tounitname=="frames") or
            (self._unit=="frames" and tounitname=="timecode") # <-- no timecode type specified, should be okay
            ):
            return self._val

        if (self._unit in ('frames','timecode') and tounitname=='seconds'):
            log.debug("unit conversion assuming 30fps")
            return self._val/30
        if (self._unit=='seconds' and tounitname in ('frames','timecode')):
            log.debug("unit conversion assuming 30fps")
            return self._val*30
        
        if self._unit=='timecode' and tounitname=='seconds':
            return self.as('frames').convert(tounitname)
        return super(Duration,self).convert(tounitname)

    def step(self):
        """currently used by channel.py. i'd like to get rid of it"""
        if self._unit in ('frames','timecode'):
            return Duration("1fr")
            

class Rate(Unum2):
    """
    Unum2 can't yet form new units out of 1/Duration, so we have separate rate units.
    """
    units={'hertz':(1.0,('Hz','hz','fps')),
           'kilohertz':(10**3,('kHz','khz')),
           'megahertz':(10**6,('mHz','mhz')),
           }


def decodeunit(input,**kwargs):
    """returns a unit object by attempting to parse
    the input with various unit types."""
    v=None
    for t in Rate,Duration,Length,Timestamp,float:
        try:
            v=t(input,**kwargs)
            break
        except ValueError:
            pass
    log.debug("decoded '%s' to %r" % (input,v))
    return v


try:
    import psyco
    #psyco.bind(Unum2)
except ImportError:
    pass

if __name__=='__main__':
    logging.basicConfig()
    log.setLevel(logging.INFO)
    
    CM=Length("1cm")
    x=Length("5m")
    y=5*CM
    print x.as("cm")
    print x.as("m")+y.as("m")
    print x+y,x*4,-x+y*3-x*2
    x+=5*CM
    print x+y,x*4,-x+y*3-x*2
    
    print "\ninput_________________________ representation__________________________ string form_________"
    for i in "55px","30000/1001fps","15Hz","1987-08-09T17:59:15-07:00","00:04:00:02":
        print "%-30s %-40s %-25s" % (i,`decodeunit(i)`,str(decodeunit(i)))
    print ""
    print "add timecodes and frames: ",Duration("00:00:00:05")+Duration("00:00:00:29")+Duration("5fr")
    print "cmp test",Duration("00:00:00:40")==Duration("00:00:01:10")

    print x/y,x/2,3/x,x/x
    print `Duration("00:05:02:11")`
    print Duration('00:00:00:05')==Duration(5,'fr')
    print Duration('00:00:00:05')==Duration(5.0,'fr')
    print Duration('00:00:00:05')==Duration(5.1,'fr')
    print Duration("37fr")/Duration("83fr")
