from __future__ import division, nested_scopes
from cuisine.shownodes.effect import Effect,register
import logging
log=logging.getLogger("effect")
import new,time
from cuisine.effects import matte
from cuisine.utils import ClassyDict
from cuisine.player import Decoder
from cuisine import pyxbuf

"""
the plan:

dv on disk -> mem cache -> decode -> composite --+-> display 
                                                  \
                                                 recode
             	                                    \ 
					             \--> new compressed dv file

"""

# 3 decoders, since we need all their pixel buffers active at once. yes, there
# would be better ways to achieve this.
yuvdecoders=dict([(name,Decoder("yuv")) for name in ['fg','bg','matte']])

class Comp(Effect):
    """
    takes 2 video inputs 'fg' and 'bg' and mixes them with

      out=fgScl*fg + bgScl*bg

    where the scl are curves read off the fg and bg tracks. they
    default to 1. SCL ARE NOT IMPLEMENTED

    If you also supply a 'matte' input, this formula is used:

      out=fgScl*(1-matte)*fg + bgScl*(matte)*bg

    the matte value is the luma from the matte video, normalized to go
    from black=0 to white=1. scl factors are as before- NOT IMPLEMENTED.
    
    """
    def process(self,t,fraction,component,inputs):
        global tmp
        t1=time.time()
        if component!='video':
            log.warn("requested '%s' from an effect that doesn't support it"%component)
            return None

        #for samples,scl in zip(self.framesforinputs(t,'audio',inputs.mix),
        #                       self.curvesforinputs(t,'audioLevel',inputs.mix)):
        

        inputdata=ClassyDict() # name (fg,bg,[maybe matte]) : pixelstring (yuv data)

        for inputname in 'fg','bg','matte':
            if inputname not in inputs:
                if inputname!='matte':
                    log.warn('comp effect needs fg and bg inputs. %s missing'%inputname)
                    return None
                else:
                    inputdata.matte=None
                    continue # ok to have no matte
            
            frame=self.framesforinputs(t,'video',inputs[inputname])[0]
            if not frame:
                if inputname!='matte':
                    log.warn("comp effect needs a frame on %s input"%inputname)
                    return None
                else:
                    inputdata.matte=None
                    continue # ok to have no matte
            yuvdecoders[inputname].decode(frame)
            inputdata[inputname]=yuvdecoders[inputname].pixels

        # now we have inputdata.fg and inputdata.bg, and maybe inputdata.matte (else it's None)

        bufsize=720*480*2
        newbuf=pyxbuf.getbuffer(bufsize) # the free for this is in player.py
        t2=time.time()
        matte.matte_effect(newbuf,
                           inputdata.fg,
                           inputdata.bg,
                           inputdata.matte or inputdata.fg,
                           inputdata.matte is not None,
                           bufsize//2)
        log.debug("intro work %.2fms; matte effect took %.2fms"%((t2-t1)*1000,(time.time()-t2)*1000))
        f=ClassyDict() # interim 'string' used for holding frames. player will recognize the special attr below
        f.alreadydecoded=newbuf # player will notice this and use it instead of decoding. then it'll be del'd.
        return f


register(Comp)

if 0:
    try:
        import psyco
        psyco.bind(Comp)
    except ImportError:
        pass
