#!/usr/bin/python

# progress.py
# Based on original code by <drewp (at) bigasterisk (dot) com>
#
# this version is slightly revised from the "official" one at http://meson.dyndns.org:8081/progress.py


import time,sys

class UserError(Exception):
    pass

def makeStatBar(width, maxPosition, doneChar = '=', undoneChar = '-', currentChar = '>'):
    """makeStatBar(width, maxPosition, doneChar = '=', undoneChar = '-', currentChar = '>') -> statBar callable
    
    Creates a function that will yield a progress bar.
    """
    try:
        aValue = width / float(maxPosition)
    except ZeroDivisionError:
        aValue=0
    def statBar(position, force = 0, last = ['']):
        assert len(last) == 1, UserError("Don't mess with the last parameter, bum.")
        done = int(aValue * position)
        toDo = width - done - 2
        result = "[%s%s%s]" % (doneChar * done, currentChar, undoneChar * toDo)
        if force:
            last[0] = result
            return result
        if result == last[0]:
            return ''
        last[0] = result
        return result

    statBar.__doc__ = """statBar(position, force = 0) -> '[%s%s%s]'-style progress bar

    returned string is %d characters long, and the range goes from 0..%d.
    The 'position' argument is where the '%s' will be drawn.  If force is false,
    '' will be returned instead if the resulting progress bar is identical to the
    previously returned progress bar.
""" % (doneChar * 3, currentChar, undoneChar * 3, width, maxPosition, currentChar)
    return statBar

if __name__ == '__main__':
    statbar = makeStatBar(75, 100)
    print statbar.__doc__
    for i in range(1,100):
        sys.stdout.write(statbar(i) + '\r')
        sys.stdout.flush()
        time.sleep(.1)
    print
