import sys
from cuisine import pygame_blit, libdv
import pygame, array
from cuisine import pyxbuf
from dispatch import dispatcher
import pyalsa,audioop
from pygame.locals import *

import logging
log=logging.getLogger("player")

from cuisine import dvfile

class Decoder:
    """simple dv->rgb decoder.
    this is not used by the stuff below, but it is used by whip and dvdecode

    """
    def __init__(self,colorspace="rgb"):

        # i forget which args are better. dvdecode.py which i used for making
        # frame files uses 0,1,1
        
        self.decoder = libdv.Decoder(0,1,1)#1,0,0)
        self.decoder.quality = 5
        dispatcher.connect(self.changequality, "libdv quality")
        
        self.colorspace = colorspace
        self.size = libdv.decodedframesize(self.colorspace)
        self.newpixelbuffer()

    def changequality(self, quality):
        log.debug("decoder %s to quality %s" % (id(self.decoder),quality))
        self.decoder.quality = quality
        
    def decode(self,frame):
        self.decoder.parseheader(frame)
        self.decoder.decodeframe(self.colorspace, frame, self.pixels)
        
    def rgbout(self):
        s = pyxbuf.buftostring(self.pixels, self.size)
        return s
    pixelsout=rgbout
    def newpixelbuffer(self):
        
        """if you kept the pixel buffer from the last decoding, call
        this to make a new one.  but this is a recipe for memleaks, so
        dont call it too much."""

        self.pixels=pyxbuf.getbuffer(self.size)
    
def describedecoder(decoder):
    "lists a few fields from the decoder structure (audio fields might be broken - I've only seen zeros)"
    for a in 'width','height','frame_size','quality','num_dif_seqs':
        print "decoder.%s = %s" % (a,getattr(decoder,a))

    for a in 'quality',:
        print "decoder.video.%s ="%a, getattr(decoder.video,a)

    for a in 'samples_this_frame','max_samples','quantization','frequency','num_channels':
        print "decoder.audio.%s ="%a, getattr(decoder.audio,a)


class Videoplay:
    """ the video-only functions"""
    overlay=None
    def __init__(self,decoder):

        # why on earth is this a separate decoder from the Decoder object above?
        
        self.decoder = decoder
        self.colorspace="yuv" # packed "YUY2" format: Y0,U,Y1,V
        # pixel size could be wrong in the decoder! use the init colorspace arg please

        def cq(quality,d=self.decoder):
            log.debug("main decoder %s to quality %s" % (id(d),quality))
            d.quality=quality
        dispatcher.connect(cq,"libdv quality",weak=0)
        
        self.decodedframesize=libdv.decodedframesize(self.colorspace)
        self.pixels=pyxbuf.getbuffer(self.decodedframesize)
        self.freefunc=pyxbuf.freebuffer

    def __del__(self):
        if self.pixels is not None and pyxbuf is not None:
            self.freefunc(self.pixels)
            self.pixels=None
        if log:
            log.info("freeing YUV overlay")
        if self.overlay is not None:
            pygame_blit.SDL_FreeYUVOverlay(self.overlay)
    def decodenewframe(self,frame):
        if not hasattr(frame,"alreadydecoded"):
            self.decoder.decodeframe(self.colorspace, frame, self.pixels)
            self.frametooverlay(self.pixels)
        else:
            self.frametooverlay(frame.alreadydecoded)
            # only works once! (to save ram, which is not well-managed)
            pyxbuf.freebuffer(frame.alreadydecoded)
            del frame.alreadydecoded

    def setupoverlay(self,surf):
        self.overlay = pygame_blit.createsdloverlay(720,480,"YUY2",surf)
        pygame_blit.SDL_LockYUVOverlay(self.overlay)

    def frametooverlay(self,pixelbuffer):
       ###    pygame_blit.blitbuffertosurface(pixels, win) # works, but sdl is faster
        pygame_blit.pixelstooverlay(pixelbuffer, self.overlay)

        pygame_blit.SDL_UnlockYUVOverlay(self.overlay)
        assert(pygame_blit.DisplayYUVOverlay(self.overlay,(0,0,720,480))==0)
        pygame_blit.SDL_LockYUVOverlay(self.overlay)
       

class Audioplay:
    def __init__(self,decoder=None):
        """decoder is optional, as this obj can be used just for playing
        samples gotten from elsewhere"""
        self.decoder=decoder
        self.audiobuffers=None # wait until we get our first frame
        self.audioout=None
        self.emptybuffer=1
        if self.decoder and hasattr(self.decoder, "disable_error_log"):
            self.decoder.disable_error_log()

    def samplesfromframe(self,dvframe):
        dv=self.decoder
        # need to find the num audio samples for this frame
        r=dv.parseheader(dvframe) 
        if r==0:
            "format stayed the same"
        if r==1:
            "format changed to pal"
        if r==2:
            "format changed to ntsc"

        if self.audiobuffers is None:
            # audio buffers can hold a whole frame's audio. one buffer
            # per channel. 32k audio untested
            chans = dv.get_num_channels()
            samps = libdv.DV_AUDIO_MAX_SAMPLES
            self.audiobuffers = libdv.makeaudiobuffers(chans,samps)

        bufs=self.audiobuffers

        # now we extract the audio into the buffers
        try:
            samps = dv.decodeaudio(dvframe,bufs)
        except RuntimeError, e:
            samps = 0
            print e
            return ""

        samplesl=bufs[0][:samps*2]
        samplesr=bufs[1][:samps*2]
        samples=audioop.add(audioop.tostereo(samplesl,2,1,0),
                            audioop.tostereo(samplesr,2,0,1),2)
      
        # conform everything to 48000
        if dv.get_frequency()==32000:
            # upsample to 48k
            samples,state=audioop.ratecv(samples,2,2,32000,48000,None)

        return samples[:len(samples)]

    def setup_audio(self,freq,chans):
        print "opening audio device with %s channels %shz" % (chans,freq)
        self.audioout=pyalsa.Pcm()
        self.audioout.setup(rate=freq, channels=chans, format_name="S16_LE",
                            start_threshold=3000,avail_min=2048)
    
    
    def playsamples(self,samples):
        # we're assuming that we can get the channel/freq settings
        # from the first frame only, which isn't true if the format
        # changes mid-tape
        dv=self.decoder
        if self.audioout is None:
            if dv:
                chans,freq = (dv.get_num_channels(),
                              dv.get_frequency())
            else:
                log.info("no dv decoder - using default audio settings")
                chans,freq = 2,48000

            # always use 48k; samplesfromframe will resample to this
            chans, freq = 2, 48000
            self.setup_audio(freq,chans)
        try:
            self.audioout.writei(samples,len(samples)/2/2)
        except pyalsa.AlsaError,e:
            print str(e)
            #self.audioout.reset()

            
    def samplesleft(self):

        """returns the number of samples left to be played on the
        hardware.

        note that calling this makes my sblive card under alsa start playing
        at 1/2 speed """
        
        if not self.audioout:
            return 0
        try:
            ret=self.audioout.delay()
        except pyalsa.AlsaError,e:
            print str(e)
            #self.audioout.reset()
            ret=0
        return ret

    def drop(self):
        if self.audioout:
            self.audioout.drop()
    

def displayaudio(buf):
    """a debug function that shows the first 32 shorts of a string in hex"""
    import array
    a=array.array('h')
    a.fromstring(buf)
    print a[:10],a[-10:]

class _Player:
    def __init__(self):
        # i think it's been the incorrect 1,0,0) for a while.
        # playdv uses 1,0,0. i still don't know which is right
        self.decoder = libdv.Decoder(0,1,1) 

        self.decoder.quality = 5 # full?
        self.vid=Videoplay(self.decoder)
        self.aud=Audioplay(self.decoder)
    def setuppygame(self):
        """if you don't want to setup pygame, use this (but after
        setting SDL_WINDOWID, if you're hoping to embed). this calls
        setsurface"""
        pygame.display.init() # don't use pygame.init, or you'll init sound too!
        SCREEN_RECT     = Rect(0, 0, 640,480)      # preferred window size
        WIN_STYLE       = DOUBLEBUF |HWSURFACE    # preferred window style | FULLSCREEN
        depth = pygame.display.mode_ok(SCREEN_RECT.size, WIN_STYLE, 32)
        sdlwin = pygame.display.set_mode((640,480))
        self.setsurface(sdlwin)

    def setsurface(self,surf):
        """an overlay will be created on the given pygame surface, and
        the frames will be drawn on it"""
        self.vid.setupoverlay(surf)

    def audiofromframe(self,frame):
        return self.aud.samplesfromframe(frame)
    
    def playvideoframe(self,frame):
        """sends a video frame only to the display"""
        self.vid.decodenewframe(frame)

    def playaudioframe(self,frame):
        """sends just the audio from a frame to the audio system"""
        self.aud.playsamples(self.aud.samplesfromframe(frame))
    def playaudioframedata(self,samples):
        """this is for already-decoded audio data, which should be a string
        in the appropriate format (probably AFMT_S16_LE stereo). length is
        variable."""
        self.aud.playsamples(samples)

    def playdvframe(self,frame,wait=None):
        """puts the given dv frame data in the (short) queue to be
        played. takes frame data or Frame object"""

        # give the frame to self.vid to decode and store
        if wait=='pre':
            self.aud.waitforplay()
        self.vid.decodenewframe(frame)
        self.aud.playsamples(self.aud.samplesfromframe(frame))
        if wait=='post':
            self.aud.waitforplay()

_master_player=None
def Player():
    global _master_player
    if _master_player is None:
        _master_player=_Player()
    return _master_player
