from __future__ import nested_scopes
import logging
log=logging.getLogger("deck")
import sys,fcntl,commands,time,os,struct
import select
from fcntl import ioctl, fcntl, F_SETFL, F_GETFL
from dispatch import dispatcher

try:
    import avc
except ImportError:
    class _:
        pass
    avc=_()
    del _


# this file got lost and has been decompyled from .pyc on 2003/6/6

class Dv1394device:
    def __init__(self, devicename = '/dev/dv1394'): # watch out - devicename might get inserted in shell command!
        self.dev = self.getdevice(devicename)
        self._frameaccum = None

    def havedevice(self):
        return (self.dev is not None)



    def getframe(self):
        """returns a string of frame data if a complete one is
        available, else returns None.  might raise an exception in the
        case of a severe io problem."""
        framebuf = None
        if (not self.havedevice()):
            return None
        if (self._frameaccum is not None):
            log.debug('partial frame waiting, read more data')
        try:
            framebuf = os.read(self.dev, 120000) # nonblocking read
        except (OSError, IOError), e:
            print 'still error',e
            if (e.args[1] == 'Resource temporarily unavailable'):
                # ok, this means there are no frames ready
                pass
            else:
                raise

        # if we read at an unlucky time, we might not get a
        # complete 120000 bytes, so there's an accumulator to
        # gather up an entire frame
        if (framebuf and (self._frameaccum or (len(framebuf) != 120000))):
            log.debug('accum section')
            if (self._frameaccum is not None):
                # append
                self._frameaccum += framebuf
            else:
                # start new accumulator
                self._frameaccum = framebuf
            log.debug(('accumulating: accum has %s bytes, framebuf has %s' % ((len(self._frameaccum) - len(framebuf)),
             len(framebuf))))
            l = len(self._frameaccum)
            if (l == 120000):
                framebuf = self._frameaccum
                self._frameaccum = None
                log.debug('partial frame is complete - discarding anyway')
                framebuf = None
            elif (l > 120000):
                log.error(('read %i bytes (more than a frame) - discarding' % l))
                framebuf = None
                self._frameaccum = None
            elif (l < 120000):
                log.debug('accumulating frame data')
                framebuf = None
        return framebuf



    def getdevice(self, devicename):
        """do any settings on the filehandle to turn up buffers,
         these are currently rather broken in
        dv1394; it's hard to get the settings, change n_frames, and then
        init with new settings.

        """

        def opendevice(devicename):
            try:
                dev = os.open(devicename, (os.O_RDONLY | os.O_NONBLOCK))
                log.info(('opened %s successfully' % devicename))
            except IOError, e:
                err = e.args[1]
                if (err == 'Device or resource busy'):
                    log.error('dv1394 device is busy; in use by:')
                    log.error(commands.getoutput(('fuser %s' % devicename)))
                else:
                    log.error(('trouble opening dv1394: %s' % e))
                return None
            return dev



        def closedevice(dev):
            os.close(dev)



        def setupdv1394_init(dv1394_init, n_frames):
            """modifies the n_frames field of the given dv1394_init
            structure.  structure contents copies from dv1394.h, SVN
            version 729 (2003/6/6), abbreviated here (more comments in
            dv1394.h):

            struct dv1394_init {
                    unsigned int api_version;
                    unsigned int channel; /* isochronous transmission channel to use */
                    unsigned int n_frames; /* 2..DV1394_MAX_FRAMES */
                    enum pal_or_ntsc format;
                    unsigned long cip_n;
                    unsigned long cip_d;
                    unsigned int syt_offset;
            };

            also stripped of comments (original is in dv1394.h):
            struct dv1394_status {
                    struct dv1394_init init;
                    int active_frame;
                    unsigned int first_clear_frame;
                    unsigned int n_clear_frames;
                    unsigned int dropped_frames;
            };



            """
            fields = list(struct.unpack(('IIIiLLi' + 'iIII'), dv1394_init))
            log.info(('replacing dv1394 n_frames=%i with %i' % (fields[2],
             n_frames)))
            log.debug('dv1394_init: %r', fields[:7])
            fields[2] = n_frames
            return struct.pack('IIIiLLi', *fields[:7])

        # this enum is copied from dv1394.h, SVN version 551
        (DV1394_INVALID, DV1394_INIT, DV1394_SHUTDOWN,
        DV1394_SUBMIT_FRAMES, DV1394_WAIT_FRAMES,
        DV1394_RECEIVE_FRAMES, DV1394_START_RECEIVE,
        DV1394_GET_STATUS,) = range(0, 8)
        
        dev = opendevice(devicename)
        if (dev is None):
            return None
        if ioctl(dev, DV1394_INIT):
            raise RuntimeError('DV1394_INIT ioctl failed')
        arg = ioctl(dev, DV1394_GET_STATUS, struct.pack(('IIIiLLi' + 'iIII'), *([0] * 11)))

        # flush any frames sitting in the buffer
        for i in range(0, 30):
            try:
                dev.read(120000)
            except:
                pass

        return dev



class Deck:

    """a deck on the firewire network, controlled by AVC. frames come
    in on dv1394.  frame-specific methods (such as getrecorddate) use
    the last-received frame. use addnewframecb() to add a callback
    function that shall receive each new frame.

    this object is the sole interface to the camera hardware. if
    someday you wanted to handle multiple decks at once, it would
    probably be best to make an i/o object that can multiplex the
    requests to and from multiple Deck objects."""
    
    dvdecode = None

    def __init__(self):
        # (user should be allowed to choose a device)
        (self.handle, self.device,) = avc.setup()
        log.info(('queried AVC device on %s, %s' % (self.handle,
         self.device)))
        if (self.device == -1):
            log.warn("there's no AVC decide available - proceeding witoout AVC control or ATN reading")
            self.device = None
        self.dv1394 = Dv1394device()
        self.atn = None
        self.recorddate = None
        self.newframecbs = [] #  callbacks to call with every new frame we read
        self.frame = None # the frame object that holds the last-seen frame
        self.earlyreturn = 0 # did the last poll() return early to allow for an autosplit file change


    def getATN(self):
        """returns absolute track number (integer), which always comes from avc"""
        return self.atn



    def getrecorddate(self):
        """returns record date as a string like 0000-00-00T00:00:00-00:00"""
        return self.frame.recorddate



    def getrecorddatedict(self):
        """returns last-seen record date as a dictionary of values suitable for use in name schemes
        or {} if we have no record date"""
        # since avifile.i doesn't even try to give us the values, we disassemble its formatted string
        d = self.getrecorddate()
        if (d is None):
            return {}
        return {'recyear': int(d[0:4]),
         'recmonth': int(d[5:7]),
         'recday': int(d[8:10]),
         'rechour': int(d[11:13]),
         'recminute': int(d[14:16]),
         'recsecond': int(d[17:19])}



    def addnewframecb(self, cb):
        """set a callback to be called with each new dv frame"""
        self.newframecbs.append(cb)



    def delnewframecb(self, cb):
        """remove a callback that you added in before"""
        self.newframecbs.remove(cb)



    def isnewrecording(self):
        """is the last frame read the start of a new recording? (in which case the
        callback wouldn't have happened yet, but it will happen at the beginning of
        the next poll.

        but there's more: this function will return true only after
        one poll (even if the frame is IsNewRecording for the next few
        polls). so you can make autosplit decisions based on this
        return value turning true exactly once. the poll() function
        arranges for this to be true before the new frame is send to
        the newframe callbacks. see poll() for more details.
        
        """
        return self.earlyreturn



    def poll(self):
        """watch the firewire for dv frames OR poll with avc queries. if there
        is a new frame, the frame is returned (as a Frame object). otherwise
        None is returned. if the frame is found to be the start of a new
        recording, poll() returns right away. but on the next call to poll()
        after a new-record frame, that frame is sent to the callback before
        anything else."""
        if self.dv1394.havedevice():
            # check for an early-abort case (checking the frame is not sufficient
            # since it might not change for a few poll() calls)

            if (self.earlyreturn and (len(self.newframecbs) > 0)):
                log.debug('deck had returned early, catching up on callbacks now')
                for cb in self.newframecbs:
                    cb(self.frame)

                self.earlyreturn = 0
            framebuf = self.dv1394.getframe()
            if framebuf:
                self.frame = framebuf
                
                # note that self.atn does not get set if we're reading from frames

                # if this is start-of-recording, return early
                # and do the callback when we get called next time (see above)
                if dvframe.isnewrecording(self.frame):
                    log.debug('deck found new recording- returning early')
                    self.earlyreturn = 1
                    dispatcher.send('new recording')
                    return self.frame
                for cb in self.newframecbs:
                    cb(self.frame)

                return self.frame
            elif (self.device is not None):
                # ask the deck for timecode and atn via AVC
                #these could hang!
                #self.timecode=avc.avc1394_vcr_get_timecode(self.handle,self.device)
                self.atn = avc.avc1394_vcr_absolute_track_number(self.handle, self.device)
                if (self.atn == -1):
                    self.atn = None
        return None



    def command(self, cmd):
        # this calls any of the avc1394_vcr_<some command with no args>() functions
        if (cmd in ['play', 'reverse', 'stop', 'rewind', 'forward',
         'pause', 'record', 'eject', 'next', 'next_index', 'previous',
         'previous_index']):
            log.info('Sending avc command: %s', cmd)
            getattr(avc, ('avc1394_vcr_%s' % cmd))(self.handle, self.device)
        else:
            raise ValueError(("'%s' is not a valid vcr command" % cmd))



