
# todo:
#  gently disable logging if there's no logging module
#  testing, including handling files that change mid-scan; special/unreadable files
#  implement the recursive mode

import logging
log = logging.getLogger("dirwatch")
import os
from glob import glob

class dirwatch:
    """A dirwatch object watches one directory and can report on
    new/changed/deleted files. No callbacks are used; the caller polls
    this object for changes. A glob pattern can be provided, in which
    case dirwatch will only consider files that match. If you want a
    system that notifies you when files change, consider 'fam'.

    recursive=1 means we'll find filenames like 'find' would

    filenames are returned relative to the base path
    """

    def __init__(self,basepath,pattern=None,recursive=0):
        log.debug("new dirwatch created for directory %s" % basepath)
        self.basepath=basepath
        self.pattern=pattern
        self.recursive=recursive

        self.mtimes={} # path:mtime

    # these two methods contain the only system calls
    def getmtime(self,file):
        log.debug("stat'ing %s" % file)
        try:
            return os.stat(os.path.join(self.basepath,file))[8]
        except Exception,e:
            log.debug("stat failed for %s: %s" % (file,e))
            return None
    def getallfiles(self):
        """this does a path walk, and is NOT TESTED TO DO THE RIGHT
        THING in all file layout situations"""
        abslen=len(os.path.abspath(self.basepath))
        allfiles=[]
        def visit(arg,dirname,names):
            dirname=os.path.abspath(dirname)
            for f in names:
                f=os.path.join(dirname,f)
                if os.path.isfile(f):
                    f=f[abslen+1:] # strip basepath
                    allfiles.append(f)
            # empty directories are included
            if len(names)==0:
                allfiles.append(dirname[abslen+1:]+"/")
        os.path.walk(self.basepath,visit,None)

        return allfiles
        
        if self.pattern is not None:
            return glob(os.path.join(self.basepath,self.pattern))
        else:
            return os.listdir(self.basepath)
        
    def checkmtime(self,path):
        """checks a file against the mtimes cache; returns whether the
        file stat is different from (or wasn't in) cache"""
        mtime = self.getmtime(path)
        if self.mtimes.get(path,0)!=mtime:
            if path in self.mtimes.keys():
                log.debug("%s mtime changed to %s from %s" % (path,mtime,self.mtimes[path]))
            else:
                log.debug("%s added to mtime cache" % path)
            self.mtimes[path]=mtime
            return 1
        return 0

    def touchedfiles(self):
        """returns list of watched files that have changed/added since
        the last call. files that have been deleted are not handled
        {well,at all}. Currently, this function stat()s the plugin
        directory and all known files in it, *every time*. """
        changedfiles=[]

        # first scan the known files
        for f in self.mtimes.keys():
            if f==self.basepath:
                # don't scan the directory yet
                continue
            try:
                if self.checkmtime(f):
                    # file was created/modified
                    changedfiles.append(f)
            except OSError:
                # file was deleted?
                log.error("file %s could not be found - dirwatch's behavior is untested" % f)
                del self.mtimes[f]

        # now check the directory itself
        if self.checkmtime("."):
            # directory itself changed - scan for new files
            files=self.getallfiles()
            for f in files:
                if f not in self.mtimes.keys():
                    log.debug("found new file %s" % f)
                    changedfiles.append(f)
                    self.checkmtime(f)

        return changedfiles

if __name__=='__main__':
    for f in dirwatch('..',recursive=1).getallfiles():
        print f
