from __future__ import nested_scopes
from cuisine.pour.dirwatch import dirwatch
import gtk,os,logging,gobject,time
from cuisine.utils import *

log=logging.getLogger("filestatview")
"""
size in megabytes

make new elements get shown

make the sort stay up to date

read file size once after file is closed (and flushed)

if an update is going to take a long time, do it in multiple polls

option to recurse (and have filenames like foo/bar)
"""

def timeformat(t):
    """returns time like 0000-00-00T00:00:00-00:00 in the current timezone"""
    (y,mo,d,h,m,s,weekday,dayofyear,dst)=time.localtime(t)
    # please handle timezone using time.timezone,time.altzone,time.daylight
    tzh=8; tzm=0 # my timezone, which is negative
    return "%(y)04i-%(mo)02i-%(d)02iT%(h)02i:%(m)02i:%(s)02i-%(tzh)02i:%(tzm)02i" % locals()


class Easyclistview_treeview:
    """wrapper for TreeView with ListStore. """
    def __init__(self,existingtreeview=None,selectmode=gtk.SELECTION_EXTENDED):
        if existingtreeview is not None:
            self.view=existingtreeview
        else:
            self.view=gtk.TreeView()
            # you'd have to get this and pack/show it somehow
        self.view.set_headers_clickable(1)
        self.view.get_selection().set_mode(selectmode)

    def setupcolumns(self,columns):
        """takes a list of title,type pairs, for example:
                (('Filename',gobject.TYPE_STRING),
                 ('Modified',gobject.TYPE_STRING),
                 ('Size',gobject.TYPE_UINT))
        """
        
        self.model=apply(gtk.ListStore,map(lambda x: x[1],columns))
        self.columnnames=[x[0] for x in columns]
        self.view.set_model(self.model)

        # this is the model data, maintained in parallel since i can't access the model
        self.data=[] # list of rows, each row has columns of values
        
        renderer=gtk.CellRendererText()
        renderer.set_property('font','7')

        for colname,colnum in zip(map(lambda x:x[0],columns),range(0,len(columns))):
            col=gtk.TreeViewColumn(colname,renderer,text=colnum)
            col.set_reorderable(1)
            self.view.append_column(col)

    def sortby(self, colnum_or_name, direction="ascending"):
        if not isinstance(colnum_or_name,int):
            colnum_or_name=[i for i,x in enumerate(self.columnnames) if x==colnum_or_name][0]
        self.model.set_sort_column_id(colnum_or_name,direction)

    def getselected(self,colnum):
        """returns list of selected entries; the list will contain the
        string in the given column"""
        sel=self.view.get_selection()
        selectediters=[]
        sel.selected_foreach(lambda model,path,iter,data: selectediters.append(iter),None)
        return [self.model.get_value(iter,column=0) for iter in selectediters]

    def addrow(self,*columnvalues):
        """appends a row with the given column texts (which match the
        columns you setup with earlier. hopefully you've cast your
        columns to the right types. returns the row number"""
        iter=self.model.append()
        args=[]
        for i,v in enumerate(columnvalues):
            args+=[i,v]
        self.model.set(iter,*args)
        return 0 #?
    def iterforrow(self,rownum):
        # this fails, the null-ok on the first arg doesnt seem to work
        ##iter=self.model.iter_nth_child(None,rownum)
        
        iter=self.model.get_iter_root()
        for i in range(rownum):
            iter=self.model.iter_next(iter)
        log.debug("row %s -> iter %r" % (rownum,iter))
        return iter
    def setrow(self,rownum,*columnvalues):
        """like addrow, but replaces the values in an existing row"""
        iter=self.iterforrow(rownum)
        args=[]
        for i,v in enumerate(columnvalues):
            args+=[i,v]
        self.model.set(iter,*args)
    def findrow(self,colnum,value):
        """returns the first matching row number where the model row
        contains the given value at the given column number. returns
        None for not found"""
        iter=self.model.get_iter_root()
        while 1:
            try:
                1/0
            except:
                break
        return 0

    def removerow(self,rownum):
        iter=self.iterforrow(rownum)
        self.model.remove(iter)
    def clearmodel(self):
        self.model.clear()
    def scrollto(self,rownum):
        "makes the given row visible in the UI"
#            self.view.scroll_to_cell(self.model.get_path(iter),
#                                     None, # coluimn
#                                     1, # use align
#                                     1,1)

class Easyclistview_clist:
    def __init__(self,existingtreeview=None,selectmode=gtk.SELECTION_EXTENDED):
        if existingtreeview:
            raise ValueError("this is the CList version; can't use your existing Treeview")
        self.selectmode=selectmode
        # defer the rest, since we put the # of columns in the constructor
    def setupcolumns(self,columns):
        """takes a list of title,type pairs, for example:
                (('Filename',gobject.TYPE_STRING),
                 ('Modified',gobject.TYPE_STRING),
                 ('Size',gobject.TYPE_UINT))
        """
        if hasattr(self,'view'):
            raise RuntimeError("CList can't be generated twice")
        self.view=gtk.CList(len(columns),[x[0] for x in columns])
        v=self.view
        v.set_selection_mode(self.selectmode)
        for i in range(self.view.columns):
            self.view.set_column_auto_resize(i,1)

        self.sortbytitles()

    def sortbytitles(self):
        """turns on click-headers-to-sort. wish i had kiwi's routines here.."""

        def handle(wid,colnum):
            self.view.set_sort_column(colnum)
            self.view.sort()
        self.view.connect("click-column",handle)

        self.view.set_auto_sort(1) # new rows enter sorted

    def sortby(self,colnum_or_name,direction="ascending"):
        pass
    def getselected(self,colnum):
         """returns list of selected entries; the list will contain the
         string in the given column"""
         return self.view.selection
    def addrow(self,*columnvalues):
        """appends a row with the given column texts (which match the
        columns you setup with earlier. hopefully you've cast your
        columns to the right types. returns the row number"""
        log.debug("adding row %r" % (columnvalues,))
        n=self.view.append(map(str,columnvalues))
        return n
    def setrow(self,rownum,*columnvalues):
        """like addrow, but replaces the values in an existing row"""
        if not 0<=rownum<self.view.rows:
            raise ValueError("row number %r is out of range" % rownum)
        for colnum,text in enumerate(columnvalues):
            self.view.set_text(rownum,colnum,str(text))
        return rownum
    def findrow(self, colnum, value):
        """returns the first matching row number where the model row
        contains the given value at the given column number. returns
        None for not found"""
        for r in range(self.view.rows):
            if self.view.get_text(r,colnum)==value:
                return r
        return None
    def removerow(self,rownum):
        self.view.remove(rownum)
    def clearmodel(self):
        self.view.clear()
    def scrollto(self,rownum):
        "makes the given row visible in the UI"
        if rownum<0:
            rownum+=self.view.rows
        self.view.moveto(rownum,0,0,0)

Easyclistview=Easyclistview_clist

class FileStatView(Easyclistview):
    def __init__(self,existingtreeview=None):
        """you can pass an optional treeview widget that we will turn into
        the fileview. otherwise, a new treeview is made in self.view"""

        Easyclistview.__init__(self,existingtreeview)

        columns=(('Filename',gobject.TYPE_STRING),
                 ('Modified',gobject.TYPE_STRING),
                 ('Size',gobject.TYPE_UINT))

        self.setupcolumns(columns)
       
#        self.view.connect("click-column",lambda w,e: self.onclickcolumn(w,e))
        self.sortby('Modified')

    def setpath(self,path):
        """reset the widget to a new directory"""
        self.path = path
        self.clearmodel()
        self.position={}
        self.dir=dirwatch(self.path,pattern=None,recursive=1)

        self.updateall()
    def getpath(self):
        return self.path
    def updateall(self):
        """finds changed files and updates them in the list - call this when you're idle"""
        for f in self.dir.touchedfiles():
            try:
                fstat=os.stat(os.path.join(self.path,f))
            except:
                log.debug('stat failed on %s' % f)
            else:
                self.updatefile(f,fstat[8],fstat[6])
        return 1
    def getselectedfilenames(self):
        "returns list of filenames of the currently selected files"
        return self.getselected(colnum=0)
        
    ###################################
    def onclickcolumn(self,wid,ev):
        log.debug("clicked new column")
        print wid,ev

    def updatefile(self,filename,newmod,newsize):
        """updates the stats for a file, appending it to the list if it's not already there"""

        filename=str(filename)
        modstring=timeformat(newmod).replace('T',' ')[:-6] # the T is ugly, as is the timezone
        newrow=(filename,modstring,int(newsize))

        r=self.findrow(0,filename)

        if r is None:
            newnum=self.addrow(*newrow)
            self.scrollto(newnum)
        else:
            self.setrow(r,*newrow)

        
if __name__=='__main__':    
    import logging
    log=logging.getLogger()
    log.setLevel(logging.DEBUG)
    h=logging.StreamHandler()
    h.setFormatter(logging.Formatter("%(levelname)-5s %(filename)s:%(lineno)d: %(message)s"))
    log.addHandler(h)

    win=gtk.Window()
    win.connect ('destroy', lambda win: gtk.main_quit ())        


    # test FileStatView
    files=FileStatView()
    files.setpath("descnodes")
    files.view.show()
    win.add(files.view)
    
    win.set_default_size (280, 250)

    win.show_all()
    gtk.timeout_add(2500,files.updateall)
    gtk.main()
