
from __future__ import nested_scopes,division
import gtk,gtk.glade,sys,gobject,os,time
from math import pi,sin,cos
import gtk.keysyms as key
from gnome.canvas import *
from cuisine.pour.filestatview import FileStatView
from cuisine.pour.dicteditor import DictEditor
from cuisine.player import Player
import pygtkutil, avifile
from cuisine.utils import *

import dispatcher
import logging
log=logging.getLogger("UI")


def affine_rotate(theta):
    # from libart_lgpl-2.3.9/art_affine.c
    s=sin(theta*pi/180)
    c=cos(theta*pi/180)
    return (c,s,-s,c,0,0)

class Timeline(Canvas):
    def __init__(self,*args,**kwargs):
        Canvas.__init__(self,*args,**kwargs)
        self.set_size_request(600,300)
    def drawgrid(self):
        font=gtk.load_font("-monotype-abadi mt condensed light-light-r-condensed-*-*-110-*-*-p-*-iso8859-15")
        font=("-monotype-arial-medium-r-normal-*-*-100-*-*-p-*-iso8859-15")
        font="arial 6"
        for x in range(0,900,12):
#            self.root().add(CanvasRect,x1=x,y1=0,x2=x+5,y2=100,fill_color='white',
#                            outline_color='black', width_units=1.0)
            t=self.root().add(CanvasText,x=x+2,y=40+10*((x%24)>10),
                              font=font,text=str(x),fill_color='black',                           
                              anchor=gtk.ANCHOR_W)
            self.l=self.root().add(CanvasLine,fill_color='black')
            self.l.set_property("smooth",1)
            self.l.set_property('points',[x,4,x,100,x+50,50,x+80,80])
        self.root().set_property('x',-400)
        
if __name__=='__main__':    
    import logging
    log=logging.getLogger()
    log.setLevel(logging.DEBUG)
    h=logging.StreamHandler()
    h.setFormatter(logging.Formatter("%(levelname)-5s %(filename)s:%(lineno)d: %(message)s"))
    log.addHandler(h)

    win=gtk.Window()
    win.connect ('destroy', lambda win: gtk.main_quit ())        

    if 0:
        # test FileStatView
        files=FileStatView()
        files.setpath("/home/drewp/projects/editor/pour")
        files.view.show()
        win.add(files.view)
    
    win.set_default_size (280, 250)

    win.show_all()
    gtk.timeout_add(2500,files.updateall)
    gtk.main()

class currentnamewidget:
    """mixin with all the current name widget controls"""
    namestate="not in use" # or "recording to" or "to change to"
    def __init__(self):
        self.templatedname=None
        dispatcher.connect(self.updatename,signal="templatedname changed",weak=1)
    def setname(self,tn):
        self.templatedname=tn
    def updatename(self):
        # the file view stays current with the path
        if os.path.isdir(self.namedict.get('path','')):
            self.fsv.setpath(self.namedict['path'])
        self.xml.get_widget("currentname").set_text(str(self.namedict))
    def on_currentname_changed(self,wid,ev):
        if not self.recorder:
            return
        if self.recorder.isrolling():
            # they're changing the name during recording.  the name
            # should change color, and the status line should announce
            # "Press enter to move to new name or ESC to cancel".  the
            # recorder only gets a move-to-new-name message if they
            # press enter. on ESC, we retrieve the name that the
            # recorder's currently using and put it back here.
            log.warn("mid-record name change not supported yet")
        else:
            self.recorder.setfile(os.path.join(self.fsv.getpath(),wid.get_text()))
    def on_currentname_key_press_event(self,wid,ev):
        if ev.keyval==key.Escape:
            log.warn("unhandled esc")
        if ev.keyval==key.Return:
            log.warn("unhandled return")
            
    def changenamestatus(self,msg,internalmsg):
        """we get this if recorder sends us a new status message. we use these
        messages to know what to do with the name widget too"""
        if internalmsg=="startrecord":
            # recording just started
            self.namestate="recording to"
        if internalmsg=="stoprecord":
            self.namestate="not in use"
        s=self.xml.get_widget("currentnamestatus")
        s.pop(0)
        s.push(0,str(msg))

    def updateschemecombo(self):
        """fill in the right choices for the template name scheme combo box"""
        self._changingmenu=1
        optmenu=self.xml.get_widget("namescheme")
        oldindex=optmenu.get_history()
        menu=gtk.Menu()
        for choice in self.namedict.gettemplatechoices():
            item=gtk.MenuItem(str(choice))
            menu.add(item)
            item.show()

        optmenu.set_menu(menu)
        optmenu.set_history(oldindex)
        del self._changingmenu

    def on_namescheme_changed(self,wid):
        log.debug("namescheme changed %s" % (wid))
        if not hasattr(self,'_changingmenu'):
            optmenu=self.xml.get_widget("namescheme")
            templ=self.namedict.gettemplatechoices()[optmenu.get_history()]
            self.namedict.choosetemplate(templ)
    

class UI(currentnamewidget):
    recorder=None
    def __init__(self):
        # the mixins
        currentnamewidget.__init__(self)

        self.idle=lambda: None # we'll call this in our idle time (along with our own stuff)

        # this builds all the gtk widgets. they can now be accessed through self.xml
        self.xml=gtk.glade.XML(localfile("pour.glade"))

        # now we connect the gtk signals to the right handlers,
        # which are mostly just methods of this class
#        mymethods=self.__dict__
        mymethods=dict([(n,getattr(self,n)) for n in dir(self)])
        # add the transport callbacks to the method table
        for cmd in 'rewind play pause stop forward'.split():
            mymethods["on_%s_clicked"%cmd]=lambda wid,cmd=cmd: self.deck.command(cmd)

        # get bound versions of some signal handlers
        mymethods["on_currentname_changed"]=lambda wid,ev=None,self=self: self.on_currentname_changed(wid,ev)
        mymethods["on_namescheme_changed"]=lambda wid: self.on_namescheme_changed(wid)
            
        self.xml.signal_autoconnect(mymethods)

        # file listing
        self.fsv=FileStatView(self.xml.get_widget("treeview1"))
        self.fsv.setpath(".")
        self.lastfsvtime=0 # time of the last update

        # varval table
        # there's a demo in there now
        old=self.xml.get_widget("varvaldemotable")
        par=old.get_parent()
        old.destroy()
        # replace it with a real VarValList widget
        self.vvl=DictEditor()
        self.vvl.show()
        par.pack_start(self.vvl)
        # put this new widget at the top:
        par.reorder_child(self.vvl,0)
        # when the name changes, the dicteditor needs to reread it
        dispatcher.connect(self.vvl.readdict,signal="templatedname changed")

        # timelines
        fulltimeline = Timeline(aa=1)
        self.xml.get_widget("fulltimeline").add(fulltimeline)
        fulltimeline.drawgrid()
        fulltimeline.show()
        self.fulltimeline=fulltimeline

        closeuptimeline = Timeline(aa=1)
        self.xml.get_widget("closeuptimeline").add(closeuptimeline)
        closeuptimeline.drawgrid()
        closeuptimeline.show()

        # video image - unsupported, causes crashes
        if 0:
            vi=self.xml.get_widget("videoimage")
            os.environ['SDL_WINDOWID']=str(pygtkutil.getxid(vi.window))
            self.player=Player()
            self.player.setuppygame()

        # when the name template changes (perhaps changed its template choices),
        # update the combo box that lists them
        dispatcher.connect(self.updateschemecombo,"templatedname changed")

    def newframe(self,frame):
        data=avifile.getframedata(frame)
        self.player.playdvframe(data)

    def replacewidget(self,xmlname,replacement):
        """cuts out that widget from the xml tree, puts in and shows the replacement widget,
        and returns the replacement"""
        raise NotImplementedError
        old=self.xml.get_widget(xmlname)
        par=old.get_parent()
        old.destroy()
        # replace it with a real VarValList widget
        self.vvl=DictEditor()
        self.vvl.show()

    # methods for hooking this ui to the other objects
    def setdeck(self,d):
        self.deck=d

        # causes crashes - see __init__
        #d.addnewframecb(self.newframe)
    def setrecorder(self,r):
        self.recorder=r
        dispatcher.connect(self.changenamestatus,"recorder status")
    def setnamedict(self,nd):
        """
        we need to know where the templatedname is

          connect the currentnamewidget to the TN 'changed' callback
            too, so the current name will always change when the TN
            changes. this still allows a little room for the user to
            edit the currentnamewidget on his/her own and save to a
            custom filename. but touching the dict in any way (such
            incr serial number) will trigger a 'changed' callback and
            the user's custom naming might get lost. the
            currentnamewidget.namestate might want to reflect that the
            name is a modified name, and be more resistant to
            recalculating it, somehow.
        """
        self.namedict=nd
        self.vvl.readdict(newdict=nd)

    def setidle(self,i): self.idle=i

    def mainloop(self):
        gtk.timeout_add(10,self.updatecurrenttimes) # call at least as often as the video framerate
        gtk.main()

    #
    # the rest are callbacks
    #
    def mainquit(*args):
        gtk.main_quit()

    def updatecurrenttimes(self,*args):
        """this is the idle call from the gtk mainloop. it's the
        master idle callback. it calls the other idle function and
        then takes care of the UI idle tasks """

        # run idle function from the main program
        self.idle()

        # update the tape position counters
        self.xml.get_widget("absolutetracknum").set_text(str(self.deck.getATN()))
        self.xml.get_widget("timecode").set_text(str(self.deck.gettimecode()))
        self.xml.get_widget("recorddate").set_text(str(self.deck.getrecorddate()))

        # we don't stat the directory every time; only if it's been
        # long enough since the last update do we rescan the file list
        if time.time()>self.lastfsvtime+.2: # <-- seconds between updates
            self.fsv.updateall()
            self.lastfsvtime=time.time()

        return 1 # so we get called again

    # handlers for the record buttons
    def on_start_record_clicked(self,*args):
        
        self.recorder.setrecording(rolling=1)
    def on_stop_record_clicked(self,*args):
        self.recorder.setrecording(rolling=0)

    # file listing handlers
    def on_delete_clicked(self,*args):
        "delete selected files"
        log.warn("should delete these files (unimplemented): %s"%self.fsv.getselectedfilenames())
        # make sure the path is right?
#        for f in self.fsv.getselected():
#            os.remove(f)
    def on_rename_clicked(self,*args):
        "rename selected files"
        log.warn("should rename these files (unimplemented): %s"%self.fsv.getselectedfilenames())
    
    def on_formatchoice_toggled(self,wid):
        self.recorder.setformat(wid.get_label().lower())


        
