cimport pyxbuf

cdef extern from "SDL.h":
    ctypedef struct SDL_Overlay:
        int w, h
        unsigned int *pitches
        unsigned char **pixels
    ctypedef struct SDL_Rect:
        int x,y,w,h
    int c_SDL_LockYUVOverlay "SDL_LockYUVOverlay" (SDL_Overlay *overlay)
    void c_SDL_UnlockYUVOverlay "SDL_UnlockYUVOverlay" (SDL_Overlay *overlay)
    int SDL_DisplayYUVOverlay(SDL_Overlay *overlay, SDL_Rect *dstrect)
    void c_SDL_FreeYUVOverlay "SDL_FreeYUVOverlay" (SDL_Overlay *overlay)
    void c_yuvtooverlay "yuvtooverlay" (object y_u_v_strings, SDL_Overlay *overlay )

cdef extern SDL_Overlay *c_createsdloverlay "createsdloverlay" (
    int width, int height, char *format_name, object screen_surface)

cdef extern from "string.h":    
    void *memcpy(void *dest, void *src, int n)

cdef class _SDL_Overlay:
    cdef SDL_Overlay *ptr
    def __new__(self, width, height, format_name, screen_surface):
        self.ptr = c_createsdloverlay(width, height, format_name,
                                      screen_surface)
    def display(self, xywh):
        cdef SDL_Rect *dstrect
        cdef SDL_Rect rect
        rect.x, rect.y, rect.w, rect.h = xywh
        dstrect = &rect
        return SDL_DisplayYUVOverlay(self.ptr, dstrect)

    def from_strings(self, y_u_v_strings):
        y,u,v = y_u_v_strings
        c_yuvtooverlay((y,u,v), self.ptr)
        
def createsdloverlay(width, height, format_name, screen_surface):
    ovl = _SDL_Overlay(width, height, format_name, screen_surface)
    return ovl

def SDL_LockYUVOverlay(_SDL_Overlay ovl):
    c_SDL_LockYUVOverlay(ovl.ptr)

def SDL_UnlockYUVOverlay(_SDL_Overlay ovl):
    c_SDL_UnlockYUVOverlay(ovl.ptr)

def SDL_FreeYUVOverlay(_SDL_Overlay ovl):
    c_SDL_FreeYUVOverlay(ovl.ptr)

def pixelstooverlay(pyxbuf.Buf pixels, _SDL_Overlay ovl):
    memcpy(ovl.ptr.pixels[0],
           pixels.buf,
           ovl.ptr.pitches[0] * ovl.ptr.h);

def DisplayYUVOverlay(_SDL_Overlay ovl, xywh):
    cdef SDL_Rect *dstrect
    cdef SDL_Rect rect
    rect.x, rect.y, rect.w, rect.h = xywh
    dstrect = &rect
    return SDL_DisplayYUVOverlay(ovl.ptr, dstrect)

    
