
  //#define DEBUG 1

#include <pygame/pygame.h>
#include <SDL.h>

SDL_Overlay *createsdloverlay( int width, int height, char *format_name, PyObject *screen_surface )
{
  SDL_Surface *screen;
  SDL_Overlay *overlay;
  Uint32 format;
  int i;

  screen = PySurface_AsSurface(screen_surface);
  if(!screen) {
    PyErr_SetString(PyExc_TypeError, "1st argument is not a Surface");
    return NULL;
  }

  format=0;
#define FORMAT_SET(x) if( !strcmp(format_name,#x) ) format=SDL_##x##_OVERLAY;
  FORMAT_SET(YV12) /* Planar mode: Y + V + U */
  FORMAT_SET(IYUV) /* Planar mode: Y + U + V */
  FORMAT_SET(YUY2) /* Packed mode: Y0+U0+Y1+V0 */
  FORMAT_SET(UYVY) /* Packed mode: U0+Y0+V0+Y1 */
  FORMAT_SET(YVYU) /* Packed mode: Y0+V0+Y1+U0 */
  if( !format ) {
    PyErr_SetString(PyExc_TypeError, "format name must be supported by sdl: YV12 IYUV YUY2 UYVY YVYU");
    return NULL;
  } 

  overlay=SDL_CreateYUVOverlay(width, height, format, screen);

#ifdef DEBUG
  fprintf(stderr,"Created %dx%dx%d %s %s overlay\n",overlay->w,overlay->h,overlay->planes,
	  overlay->hw_overlay?"hardware":"software",
	  overlay->format==SDL_YV12_OVERLAY?"YV12":
	  overlay->format==SDL_IYUV_OVERLAY?"IYUV":
	  overlay->format==SDL_YUY2_OVERLAY?"YUY2":
	  overlay->format==SDL_UYVY_OVERLAY?"UYVY":
	  overlay->format==SDL_YVYU_OVERLAY?"YVYU":
	  "Unknown");
  for(i=0; i<overlay->planes; i++) {
    fprintf(stderr,"  plane %d (@%p): pitch=%d\n", i, overlay->pixels[i], overlay->pitches[i]);
  }
#endif

  return overlay;
}

// from pygame-1.6-overlay.patch that goes with pymedia
// call with a tuple of strings, like from the output of pymedia.video.vcoder.Decoder.decode
void yuvtooverlay( PyObject *y_u_v_strings, SDL_Overlay *overlay )
{
  int ls_y, ls_u, ls_v, y;
  unsigned char *dst_y, *dst_u, *dst_v, *src_y, *src_u, *src_v;
  if(!PyArg_ParseTuple(y_u_v_strings, "s#s#s#", 
		       &src_y, &ls_y, &src_u, &ls_u, &src_v, &ls_v))
    // raise exception here?
    return;

  // No clipping at this time( only support for YUV420 )
  dst_y = (char*)overlay->pixels[ 0 ];
  dst_v = (char*)overlay->pixels[ 1 ];
  dst_u = (char*)overlay->pixels[ 2 ];
  for (y=0; y< overlay->h; y++)
    {
      memcpy( dst_y, src_y, overlay->w );
      
      src_y += ls_y / overlay->h;
      dst_y += overlay->pitches[ 0 ];
      
      if (y & 1) {
	src_u += ( ls_u* 2 )/overlay->h;
	src_v += ( ls_v* 2 )/overlay->h;
	dst_u += overlay->pitches[ 1 ];
	dst_v += overlay->pitches[ 2 ];
      }
      else
	{
	  memcpy( dst_u, src_u, ( ls_u* 2 )/overlay->h );
	  memcpy( dst_v, src_v, ( ls_v* 2 )/overlay->h );
	}
    }  
}


void recttest( PyObject *pyrect )
{
  GAME_Rect *gr,temprect;
  
  printf("recttest\n");
  printf("check %i\n",PyRect_Check(pyrect));

  gr=GameRect_FromObject(pyrect,&temprect);
  printf("recttest ok\n");
}

int old_DisplayYUVOverlay( SDL_Overlay *overlay, PyObject *py_dstcoord )
{
  /*
    dstcoord is a (x,y,w,h) tuple, not a Rect because i couldn't make that work
  */
  GAME_Rect *gr,temprect;
  SDL_Rect *dstrect,rect;
  
#if 1
  //printf("displayyuvoverlay\n");
  PyArg_ParseTuple(py_dstcoord,"iiii",&rect.x,&rect.y,&rect.w,&rect.h);
    /*  gr=GameRect_FromObject(py_dstrect,&temprect);
  if(!gr) {
    PyErr_SetString(PyExc_TypeError, "Invalid recstyle argument");
    return -1;
  }
  printf("gr\n");
  dstrect = (SDL_Rect*)gr;
  if( dstrect==NULL ) {
    PyErr_SetString(PyExc_TypeError, "2nd argument is not a pygame rect");
    return -1;
    }*/
    dstrect=&rect;
    //printf("rect was at %i,%i,%i,%i\n",dstrect->x,dstrect->y,dstrect->w,dstrect->h);
#else
  rect.x=rect.y=0; rect.w=640; rect.h=480;
  dstrect=&rect;
#endif
  
  return SDL_DisplayYUVOverlay( overlay, dstrect);

}

/* from SDL */
extern  void SDL_UnlockYUVOverlay(SDL_Overlay *overlay);
extern int SDL_LockYUVOverlay(SDL_Overlay *overlay);
 extern void SDL_FreeYUVOverlay(SDL_Overlay *overlay); 


void blitbuffertosurface( unsigned char *pixels, PyObject *dest_surface)
{
  // everything's fixed at 720x480

  SDL_Surface *src, *dest;
  SDL_Rect rect;

  dest = PySurface_AsSurface(dest_surface);
  if(!dest) {
    PyErr_SetString(PyExc_TypeError, "2nd argument is not a Surface");
    return;
  }

  src=SDL_CreateRGBSurfaceFrom( pixels, 720, 480, 32, 720*2, 
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
				0xff, 0xff<<8, 0xff<<16, 0
#else
				0xff<<24, 0xff<<16, 0xff<<8, 0 // untested
#endif
				);

  if(!src) {
    PyErr_SetString(PyExc_TypeError, "could not create SDL surface from buffer");
    return;
  }
  rect.x=rect.y=0;
  rect.w=720;
  rect.h=480;
  SDL_BlitSurface( src, &rect, dest, &rect);

  SDL_FreeSurface(src);

}

