from __future__ import division, nested_scopes, generators
from cuisine.xmlnodebase import xmlnodeclass,xmldocfile,collectiveelement
import sys,os,new,bisect
from dispatch import dispatcher
import xml.dom.minidom as minidom
from cuisine.desctypes import *
from cuisine.utils import *
from cuisine.external.dictmixin import DictMixin
import audioop

from cuisine import shownodes
import sys

import logging
log=logging.getLogger("shownodes")


class EffectBrickMixin:
    def inputs(self):
        """inputs are other tracks that go into named arguments for this
        effect. clip and show bricks use the <source> element, not this one.

        returns a dict-like object with keys of input names
        and values that are lists of inputs. inputs xml
        looks like this:

        <inputs>
          <input name='A'> <!--input name must be a valid python arg name-->
            <track>a2</track> <!--(these are track names, which don't have to be valid python)-->
            <track>a3</track>
          </input>
          <input name='B'>
            <track>1</track>
          </input>
        </inputs>
        
        """
        class InputDict(DictMixin):
            """clearly, this should be factored out with the similar code
            in the Edit brick (although we allow multiple values per key, so
            all 'items' are going to be tuples."""
            def __getitem__(self, key):
                raise NotImplementedError
            def __setitem__(self, key, value):
                raise NotImplementedError
            def __delitem__(self, key):
                raise NotImplementedError    
            def keys(self):
                raise NotImplementedError
        
        return InputDict(self)


    def inputstruct(self):
        """returns a ClassyDict like { 'inputname1' : (track1,track2,...), ... }.
        this doesn't change over time (unless the user alters the effect inputs)"""
        inputs=ClassyDict()

        # *good lord! get this xml parser out of here!*
        inputsblock=self.getChildrenElementsByTagName('inputs')[0]
        for inputblock in [n for n in inputsblock.childNodes if
                           n.nodeType==minidom.Node.ELEMENT_NODE]:
            inputname=inputblock.getAttribute("name")
            inputs[inputname]=[]
            for trackelem in [n for n in inputblock.childNodes if
                              n.nodeType==minidom.Node.ELEMENT_NODE]:
                trackname=self._getorsettextfornode(trackelem)
                track=self.parentbytype(shownodes.Show).trackbyname(trackname)
                inputs[inputname].append(track)

        return inputs
    
    def effect(self,newval=None):
        """the effect name is a module name in the plugins directory."""
        return self._getorsetstringnode("effect",newval)
    def __str__(self):
        return "[%s]" % self.effect()

    def tbd_getframe(self,t,component):
        # *cough*

        module=__import__("cuisine.effects."+self.effect(),
                             globals(), locals(),
                             [self.effect().capitalize()])
        effecttype=vars(module)[str(self.effect().capitalize())]
        effectobj=effecttype(self)

        if 1 or not hasattr(self,'cacheinputs'):
            self.cacheinputs=self.inputstruct()
        inputs=self.cacheinputs
        
        fraction=0

        return effectobj.process(t,fraction,component,inputs)
        
class Effect:
    """interface for effects. in the pour/effects directory there
    are files that contain classes for us to import. the classes
    are subclasses of this Effect interface.

    i think this object will get instantiated once per brick of the
    particular effect subclass (so it's okay to leave state in
    it). but it probably won't get instantiated before the effect is
    needed. __init__ will get called before any processing is done.

    """

    def __init__(self,brick):
        self.brick=brick
        
    def framesforinputs(self,t,component,inputtracks):
        """convenience method to fetch all the frames for an input
        (given as a list of Tracks). the return is a list of frame
        data elements for the given list of tracks (with None values
        for tracks that didn't provide output). for example:

        A_frames=self.framesforinput(t,'video',inputs.A)
        """
        return [self.brick.getothertrackframe(track,t,component) for track in inputtracks]

    def curvesforinputs(self,t,param,inputtracks):
        """convenience method fetches the channel values for the given
        param name.  a list is returned with the param values for each
        of the inputtracks (with None values for tracks that didn't
        have any value for that param)
        """
        ret=[]
        for b,bt in [self.brick.getothertrackbrick(track,t,returntime=1) for
                     track in inputtracks]:
            curvevalue=None
            if b:
                chans=b.channels()
                if chans and param in chans.allparams():
                    chan=chans.byparamname(param)
                    curvevalue=chan.getchannelvalue(bt)

            ret.append(curvevalue)
        return ret
    
    def process(self,
                t,         # requested time (range?), relative to effect origin
                fraction,  # 0 at effect intime to 1 effect outtime
                component, # which component should be returned ('audio','video')
                inputs,    # { 'inputname1' : (track,track...), ... }, as a ClassyDict
                ):
        """
        this is called to request an effect calculation. for now, the
        time t is the requested frame, although someday it may become
        a time range.

        the return should be a suitable video frame or audio data, according to the
        requested component. returning None means 'there was not enough data to
        calculate the effect, or something else went wrong'.
        
        """
        return None

    def getinputs(self,*args):
        """tbd: this will be how the effect tells what args are valid"""
        pass

def register(*args):
    pass
