from __future__ import nested_scopes,division
import gtk
import gtk.gdk as gdk
import gnome.canvas as canvas
keys=gtk.keysyms
from cuisine.shownodes import *
from cuisine.utils import *
from math import exp
from time import time
import sys
from skim.zooming import Pair
from cuisine.toss import pygtkutil
from cuisine.external import progress
from dispatch import dispatcher
import logging
log=logging.getLogger("timeline")
flog=logging.getLogger("timeline-frequent") # events that happen, say, on every mouse motion

basefont="arial 7" #6, 9
trackheight=70

from cuisine.toss.timeline import editbrickview
from cuisine.toss.timeline import brickview
from cuisine.toss.timeline import gtkutils
from cuisine.toss.timeline.moreobjs import Trackview, Gridlines, Timecursor, Wrappedtext


def multiargsmethod(fn):
    """given a method that takes self and one argument, this returns
    a new version that will take optional additional arguments and return
    a list. like this:
       def convert(self,x):
          return f(x)
       convert=multiargsmethod(convert)

       xc=convert(x)
       xc,yc,zc=conver(x,y,z)
    """
    def multiargversion(self,*args):
        if len(args)==1:
            return fn(self,args[0])
        else:
            return [fn(self,x) for x in args]
    return multiargversion

class Canvasdrawingmixin:
    """ wrapper for gtk's canvas commands (written such that a tk target
    hopefully wouldn't be out of the question)"""
    
    def makebox(self,x1=0,y1=0,x2=0,y2=0):
        return self.canvas.root().add(canvas.CanvasRect, x1=x1,y1=y1,x2=x2,y2=y2,
                                      outline_color="black",width_pixels=1)
    def makefilledrect(self,x1=0,y1=0,x2=0,y2=0):
        return self.canvas.root().add(canvas.CanvasRect, x1=x1,y1=y1,x2=x2,y2=y2,
                                      fill_color="gray")
    def makeline(self,*ptpairs):
        l=self.canvas.root().add(canvas.CanvasLine,fill_color="black")
        if not ptpairs:
            ptpairs=[0,0,0,0]
        l.set_property("points",ptpairs)
        return l

    def _convertanchor(self,prettyname):
        "takes nw n ne e center w sw s se and returns the gtk anchor constant"
        if prettyname not in 'nw n ne e center w sw s se'.split():
            raise ValueError("invalid anchor name")
        return getattr(gtk,"ANCHOR_%s" % (prettyname.upper()))
    
    def maketext(self,text="",anchor="nw",x=0,y=0):
        """
        anchor is one of 'nw n ne e center w sw s se'.split()
        
        todo:
          'fit in rectangle' option
        """
        gtkanchor = self._convertanchor(anchor)
#        return self.canvas.root().add(Wrappedtext(x=x,y=y))
    #######
        return self.canvas.root().add(canvas.CanvasText,font=basefont,
                                      text=str(text),x=x,y=y,
                                      fill_color='black',
                                      anchor=gtkanchor)
    def makewidget(self,widget,anchor="nw",x=0,y=0,width=None,height=None):
        if width is None:
            width,height=widget.size_request()
        anchor=self._convertanchor(anchor)
        return self.canvas.root().add(canvas.CanvasWidget,
                                      widget=widget,
                                      width=width,height=height,
                                      x=x,y=y,anchor=anchor)
    def removecanvasobj(self,*canvasobjs):
        for canvasobj in canvasobjs:
            if canvasobj is not None and hasattr(canvasobj,'destroy'):
                canvasobj.destroy()
    #        self.canvas.root().remove(canvasobj)


class Timelinednd(gtkutils.Dndreceiver):
    """this lets the canvas receive new clips by Dnd. i intended to
    create the clip at the first sign of motion, so you could drag the
    clip into its final place.  but bad gtk things happen when i
    retrieve the data too soon, so the clip is only created when you
    let go.

    eventually, you might be able to drag in an effect brick (or even
    an edit brick from somewhere else, i suppose). for now,
    everything's interpreted as clip bricks.

    demoplayer should be able to send a specially-encoded
    clip name that also includes start/end points.
    """
    lastpos=None
    def __init__(self,master):
        # the dnd happens on the canvas, but most of the time we call
        # methods on the Timelinewidget object itself
        gtkutils.Dndreceiver.__init__(self,master.canvas)
        self.master=master
    def motion(self,wid,context,data,x,y):
        self.lastpos=Pair(x,y)
    def leave(self,wid,context,data):
        pass
    def drop(self,wid,context,data,x,y):
        pass


    def filenames_from_data(self,data,datatype):
        
        """dnd provides a string of data, from which we have
        heuristics to find the filename that the user is trying to add
        to the show. list of filenames are returned- all are
        candidates for adding to the show."""

        usedfilenames=[] # list of files we actually add, to avoid dups

        for filename in data.split("\n"):
            filename=filename.strip()

            if filename.startswith("file://"):
                filename=filename[7:]  # mozilla's "file:///path/to/video" urls
            if filename.startswith("file:"):
                filename=filename[5:] # strip header from uri like "file:/video/clip.avi"

            if filename in usedfilenames:
                continue

            if not os.path.exists(filename):
                log.debug("file not found %r" % filename)
                continue

            usedfilenames.append(filename)

            if str(datatype)=="text/x-moz-url":
                # moz gives the link then the text, which we shall ignore
                break

        return usedfilenames
    
    def data_received(self,wid,context,data,x,y,info,dndtime):
        # get a DescFile for the dropped clip - the data might be like
        #   file:/path/to/clip[.desc] (type: text/uri-list, multiple items sep by \n)
        #   /path/to/clip[.desc]
        #   file:///path/to/file[.desc]\nfile.desc  (type: text/x-moz-url)
        #
        log.debug("received drop:")
        dumpobj(data)

        # in general, we're going to split on newlines and try to find
        # workable files on each line

        datatype=data.type
        data=data.data
        if data[1]=='\x00': # undo utf-16
            data = unicode(data, "utf-16").encode("iso-8859-1")

        try:
            for filename in self.filenames_from_data(data,datatype):

                if filename.endswith(".desc"):
                    newdf=DescFileManager().openfilebydescfilename(filename)
                else:
                    # this is old-style (we used to deal in media
                    # filenames directly).  however, someday we may
                    # want to accept media filenames and autogenerate
                    # desc files for them.
                    newdf=DescFileManager().openfileforclip(filename)

                if newdf is None:
                    log.info("couldn't open descfile for '%s' - ignoring" % filename)
                    continue
                clip=newdf.clip()

                # if we drop multiple files, they shouldn't pile up-- either
                # the track should search for free space or the clips should be stacked
                # forward in time:
                track=self.master.trackforycoord(y)
                time=self.master.timeforxcoord(x)

                log.debug("placing clip %r on track %r at %r" % (clip,track,time))

                show=self.master.show
                # make and setup a brick
                s=Source()
                s.filename(filename)
                newb=Brick(source=s,track=track,
                           origin=time, # actually, the brick's in-time should be set here, i suppose
                           intime=Duration(0,'frames'),
                           outtime=clip.essencefile().media().video().duration()
                           )

                # add it to the show model
                show.bricks().add(newb)

                # add a view (we don't need to reload the whole show just for this)
                self.master.addbrickview(newb)
                log.info("added new brick %s" % newb)
        finally:
            context.finish(gtk.TRUE, gtk.FALSE, dndtime) # success, delete-original, timestamp
            
        dispatcher.send("timeline changed")
        self.master.update()

class Timelinewidget(Canvasdrawingmixin):
    """this is the view for the shownodes.Show object (which marshals
    itself as xml)"""
    def __init__(self,master,windowsize,aa=1,):
        make_attributes_from_args('master','windowsize')
        self.canvas=canvas.Canvas(aa=aa)
        self.canvas.set_scroll_region(0,0,*windowsize)
        self.canvas.set_events(
                                gdk.EXPOSURE_MASK |
                                gdk.MOTION_NOTIFY |
                                gdk.LEAVE_NOTIFY_MASK |
                                gdk.BUTTON_PRESS_MASK |
                                gdk.POINTER_MOTION_MASK |
                                gdk.POINTER_MOTION_HINT_MASK)
        self.canvas.set_flags(self.canvas.flags()| gtk.CAN_FOCUS)
        
        pygtkutil.setcanvasclose_enough(self.canvas,4)

        self.objs=[] # all the Canvasobjs
        self.trackorder={} # Track : order it's displayed in the timeline (0 top, 1 underneath, etc)
        self.timezoom=0,1 # offset [px], scale [px/frame] (do scale first, offset is in canvas coord)

        self.setupmmbpanzoom()

        self.selectedbrickviews=[] # hack to let me do multiselect. this is a list of _brickviews_

        # make some gridlines
        # these suck right now
        #        self.objs.append(Gridlines(self))

        # one time cursor
        self.cursor=Timecursor(self)
        self.objs.append(self.cursor)


        # try to set some params from a little file left over at the last exit
        try:
            # obviously, this crap belongs in the show xml file (and shouldn't
            # be done in such a horrible way). see shutdown() for the file format.
            l=open("toss-timelineposition","rb").readlines()
            tz1,tz2,ct=map(decodeunit,map(str.strip,l))
            self.setzoom(tz1,tz2,0)
            self.cursor.changetime(ct)
        except:
            pass

        # this slows down the mouse motion events to a rate that we
        # can handle (in conjunction with the POINTER_MOTION_HINT_MASK
        # above)
        def report(wid,ev):
            if getattr(ev,'is_hint',0):
                x,y,state=ev.window.get_pointer()
            return 0
        self.canvas.connect("event",report)

        def keyboard(wid,ev):
            if ev.type!=gdk.KEY_PRESS:
                return
            # this lets you lose focus (give focus to the canvas) by
            # pressing escape
            if ev.keyval==keys.Escape:
                self.canvas.grab_focus()
            # bricks will add themselves to selectedbrickviews. this lets
            # you hit 0 and clear that.
            if ev.keyval==keys._0:
                log.info("clearing selection")
                self.selectedbrickviews=[]
                
            return 0
        self.canvas.connect('key_press_event',keyboard)
        # clicking on the canvas shall defocus anything else
#        self.canvas.connect("button_press_event",self.canvas.grab_focus)

        self.dnd = Timelinednd(self)

        # RMB jumps time to your mouse coordinate (this might go away
        # when we make the pie menu)
        def rmbzap(wid,ev):
            if ev.type==gdk.BUTTON_PRESS and ev.button==3:
                self.cursor.changetime(self.timeforxcoord(ev.x))
        self.canvas.connect('button_press_event',rmbzap)

        # track preview
        # this calls Tossapp.enteredtrack on any track we mouseover, if scrolllock is down.
        # no action on scrolllock being turned off- that's handled in Tossapp.
        def preview(wid,ev):
            if self.master.scrolllocked:#ev.state & 0x8: # alt key
                self.alerttrack(wid,ev)
            else:
                self.mousetrack=None
            return 0
        self.canvas.connect("motion_notify_event",preview)
        self.canvas.connect_after("key_press_event",self.alerttrack)

    def alerttrack(self,wid,ev):
        "sends tossapp.enteredtrack() the track at mouse y coord"
        x,y,state=ev.window.get_pointer()
        try:
            track=self.trackforycoord(y)
        except IndexError:
            return

        tossapp=self.master
        tossapp.enteredtrack(track)
        self.mousetrack=track
        
                
    def shutdown(self):
        """called when the timeline is getting closed down"""
        # see __init__ for more
        try:
            out="%s\n%s\n%s\n" % (self.timezoom[0],self.timezoom[1],str(self.cursor.time))
            open("toss-timelineposition","wb").write(out)
        except:
            pass

    def setupmmbpanzoom(self):
        """arranges to call setzoom upon mmb and c-mmb drags"""
        self._controldown=None
        def press(wid,ev):
            if ev.button==2:
                self._motionev=self.canvas.connect('motion_notify_event',motion)
                self._mmbdown=Pair(ev.x,ev.y)
                self._zoomdown=self.timezoom
                self._controldown=ev.state & gtk.gdk.CONTROL_MASK
                self._pivottime=self.timeforxcoord(ev.x) # in a zoom, this time must hold still
                
        def motion(wid,ev):
            control=ev.state & gtk.gdk.CONTROL_MASK

            # lifting or releasing the control key is just like
            # starting another drag
            if hasattr(self,'_controldown') and control!=self._controldown:
                ev2=ClassyDict()
                for attr in 'x','y','state':
                    ev2[attr]=getattr(ev,attr)
                ev2.button=2 # make it looks like a mouse event
                release(wid,ev2)
                press(wid,ev2)
            
            mousepos=Pair(ev.x,ev.y)
            delta=mousepos-self._mmbdown # mouse pixels from starting point

            if control:
                # adjust zoom, but keep the time at _mmbdown.x still
                newscl=self._zoomdown[1]*exp(delta.x/60)
                # do the zoom as a test
                self.setzoom(self._zoomdown[0],newscl,update=0)
                # see where the time went that was supposed to hold still
                drifted=self.xcoordfortime(self._pivottime)
                # zoom for real, with a correction now to keep the stationary point stationary
                self.setzoom(self._zoomdown[0]-(drifted-self._mmbdown.x),newscl,fast=1)
            else:
                # offset by the amount of the mouse movement
                self.setzoom(self._zoomdown[0]+delta.x,self._zoomdown[1],fast=1)
            
        def release(wid,ev):
            if ev.button==2:
                self.canvas.disconnect(self._motionev)
                for attr in '_motionev _mmbdown _zoomdown _controldown _pivottime'.split():
                    delattr(self,attr)
                # as we've been fast-updating, do one more to fix the display
                self.update()
        
        self.canvas.connect('button_press_event',press)
        self.canvas.connect('button_release_event',release)

    #
    # these next methods convert between canvas coords and times (and
    # after that, y coords and tracks). there are issues with
    # fractional frames vs integer timecodes, and those issues are
    # being addressed here by the extra int() roundoffs. hopefully
    # someday they can go away and be handled right (place a video
    # anywhere, and frames are selected, blended, etc, and the audio
    # can be precisely positioned)
    #
    def xcoordfortime(self,t):
        "returns the canvas x coord for the given time. you can give multiple args and get a list back!"
        return 110+int(t.getas('frames'))*self.timezoom[1]+self.timezoom[0]
    xcoordfortime=multiargsmethod(xcoordfortime)
    def timeforxcoord(self,x):
        "inverse of above"
        return Duration(int((x-110-self.timezoom[0])/self.timezoom[1]),'frames')
    timeforxcoord=multiargsmethod(timeforxcoord)
    def dtimefordxcoord(self,dx):
        """like the above, without the offset (useful for turning a
        mouse move amount to a time offset)"""
        return Duration(int(dx/self.timezoom[1]),'frames')
    dtimefordxcoord=multiargsmethod(dtimefordxcoord)
    def ycoordsfortrack(self,track):
        """returns the top and bottom canvas coords that the given
        track is supposed to fall into"""
        o=self.trackorder[track]
        return o*trackheight,o*trackheight+trackheight # top,bottom (canvas coords)
    def trackforycoord(self,y):
        "inverse of above"
        tracknum=int(y/trackheight)
        matches=[track for track,num in self.trackorder.items() if num==tracknum]
        if matches==0:
            raise ValueError("no track at that y coord")
        else:
            return matches[0]
    def xcoordsfortrack(self,track):
        """all tracks are aligned, so these are not a function of the track"""
        return 0,110,self.windowsize[0] # (xleft=0,xtitle=150,xright=800)
    def visiblexcoord(self):
        "returns the min,max visible x coords on the canvas"
        return 0,self.windowsize[0]


    def addbrickview(self,brick):
        "adds a brickview to this timeline based on the given brick"
        if brick.bricktype()=='edit':
            bv=editbrickview.Editbrickview(self,brick)
        else:
            bv=brickview.Brickview(self,brick)
        self.objs.append(bv)
    def removebrickview(self,brick):
        "remove the brickview for this brick"

        brickidx=None
        for i,o in enumerate(self.objs):
            if isinstance(o,brickview.Brickview) and o.brick==brick:
                brickidx=i
                break
        if brickidx is None:
            log.warn("there was no brickview for brick %r" % brick)
            return
        log.debug("removing obj %i, brick %s" % (brickidx,brick))
        self.objs[brickidx].withdraw()
        self.objs.pop(brickidx)
        
    def setshow(self,show):
        """note that you pass the show and not the showfile"""
        self.show=show

        # (should remove all the old tracks and bricks first!)

        # make Trackviews
        for i,t in enumerate(self.show.tracks().getall()):
            self.objs.append(Trackview(self,t))
            self.trackorder[t]=i

            
        # make Brickviews
        bricks=self.show.bricks().getall()
        statbar=progress.makeStatBar(80,len(bricks))
        for i,b in enumerate(bricks):
            sys.stdout.write(statbar(i)+'\r')
            sys.stdout.flush()
            self.addbrickview(b)
        print ""

        self.setbgcolor(self.show.bgcolor())

        self.update()

    def setbgcolor(self,col):
        """sets canvas bg color.

GdkColor bgcolor=your_color_here();
GtkStyle * style;
GtkStyle * defstyle;
GtkWidget * widget=your_widget_here();

defstyle=gtk_widget_get_default_style();
style=gtk_style_copy(defstyle);
style->bg[GTK_STATE_NORMAL]=bgcolor;
gtk_widget_set_style(widget, style);
        """
        s=self.canvas.get_style()
        s=s.copy()
        s.bg[gtk.STATE_NORMAL]=self.canvas.get_colormap().alloc_color(col)
        self.canvas.set_style(s)
    def advancetimecursor(self):
        """if we're playing back, this gets called when it's time to
        move to the next frame. we just move the time cursor, and that
        sends signals to the screens as usual"""
        self.cursor.advanceoneframe()

    def update(self,fast=0):
        "recalc all Canvasobjs who get position information from us"

        t=time()
        [o.update(fast) for o in self.objs]
        flog.debug("update took %.1f ms" % ((time()-t)*1000))
        # we only called an update because we wanted to see the result,
        # so i make the canvas draw now
#        t=time()
        self.canvas.update_now()
#        log.debug("canvas redraw took %.1f ms" % ((time()-t)*1000))
        # any mousemoves in the meantime should be ignored

    def setzoom(self,off,scl,update=1,fast=0):
        "updates canvas objects with the new zoom"
        if scl>20: # clamp the pixels/frame
            scl=20
        self.timezoom=off,scl

        if update:
            # this recalcs all the coordinates on the
            # canvas objects so they look right.
            self.update(fast)
        

if __name__=='__main__':
    commonlogsetup(["timeline-frequent"])
    from shownodes import ShowFile

    win=gtk.Window()
    win.connect ('destroy', lambda win: gtk.main_quit ())

    tl=Timelinewidget()
    win.add(tl.canvas)
    sf=ShowFile("test/myshow.show")
    tl.setshow(sf.show())

    def timechd(newtime):
        print "newtime",newtime

    def emit(wid,ev):
        print "-"*100
        sf.pprint()

    dispatcher.connect(timechd,"time cursor changed")
        
    win.connect('key_press_event',emit)

    win.set_default_size (800, 600)
    win.show_all()

    if 0:
        runstats(cmd="gtk.main()")
    else:
        gtk.main()
    
