
def setproperties(canvasobj,**props):
    for k,v in props.items():
        canvasobj.set_property(k,v)

# word-wrap function that preserves line breaks and spaces
# from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/148061
def wrap(s, width=78):
    return reduce(lambda line, word, width=width: "%s%s%s" % 
                  (line,
                   ' \n'[(len(line[line.rfind('\n')+1:]) + 
                    len(word) >= width)], word),
                   s.split(' ')
                  )

class Dndreceiver:
    """general dnd receiving, subclass to get specific dnd behavior.
    flip to the end for the methods to override. if data is not
    available for a specific call, None will be passed. remember that data.data contains
    the actual data, but other data fields might be useful to check the type.
    return values of the final 4 methods are ignored - you don't have to worry about them.

    set earlydata to 1 if you want the first motion event to request the data so the
    data can be available for all subsequent motions. this causes segfaults for me.
    
    """
    def __init__(self,widget,earlydata=0):
        "pass the widget that shall receive the dnd"

        self.earlydata=earlydata
        widget.drag_dest_set(0,[],0)#gtk.DEST_DEFAULT_ALL, target,gdk.ACTION_MOVE| gdk.ACTION_COPY)

        for sig,hand in (('drag_motion',self._motion),('drag_leave',self._leave),
                         ('drag_drop',self._drop),('drag_data_received',self._data_received)):
            widget.connect(sig,hand)

    def _getdata(self,context):
        try:
            return context._datastore
        except:
            return None
    def _motion(self,wid,context,x,y,time):
        if not hasattr(context,"_known"):
            context._known=1 # stick a flag on the context object so we only run this block once
            if self.earlydata:
                wid.drag_get_data(context,context.targets[0],time)
        context.drag_status(context.suggested_action,time)
        self.motion(wid,context,self._getdata(context),x,y)
        return 1
    def _leave(self,wid,context,time):
        self.leave(wid,context,self._getdata(context))
        if hasattr(context,"_known"):
            del context._known
        return 0
    def _drop(self,wid,context,x,y,time):
        if not self.earlydata:
            wid.drag_get_data(context,context.targets[0],time)           
        self.drop(wid,context,self._getdata(context),x,y)
        return 0
    def _data_received(self,wid,context,x,y,data,info,time):
        # we use this instead of self.data because it will probably be deleted
        # sooner than self.data in the event of a botched DnD
        context._datastore=data 
        self.data_received(wid,context,data,x,y,info,time)
        return 0

    ##########
    def motion(self,wid,context,data,x,y):
        pass
    def leave(self,wid,context,data):
        pass
    def drop(self,wid,context,data,x,y):
        pass
    def data_received(self,wid,context,data,x,y,info,time):
        pass
        

