"""
opengl display setup, zooming, flipping, selections, etc.
"""
from __future__ import nested_scopes
from OpenGL.GL import *
from OpenGL.Tk import RawOpengl
import Tkinter as tk
import sys

class opengldisplay(tk.Widget,tk.Misc):
    """
    this is based on OpenGL.Tk.RawOpengl, but it's so custom that i
    dont even subclass anymore.
    
    
    an opengl surface especially suitable for 2d drawing. no lighting
    or anything.  we keep the opengl coords the same as the screen
    coords, even when the surface resizes.

    """
    def __init__(self,master=None,cnf={},**kw):
        """

        doublebuffer is already included

        'sharelists' is an option that might help someday if we split the window
        """
        tk.Widget.__init__(self,master,'togl',cnf,kw)

        if 1:

            both=lambda *arg: [
                self.fixcoordinates(), # broken- makes nothing more show on the display
                self._redraw(),
                ]
            for evname,meth in (('Configure',both), # window resize
                                ('Visibility',self._redraw), # the amount of exposed window changed (but not by configure)
                                ('Expose',self._redraw), # the amount of exposed window increased
                                ('Map',both), # window is first created (or uniconified)
                                ('Button-3',self.fixcoordinates),
                                ):

                self.bind("<%s>"%evname,meth)

        self.initopengl()
        self.fixcoordinates()

    def initopengl(self):        
        # glDrawPixels shall do an alpha blend onto the display
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA)

    def fixcoordinates(self,*args):
        """makes the opengl coords be the same as screen coords"""

        glMatrixMode(GL_MODELVIEW)
        # left,right,bottom,top
        glLoadIdentity()
        glOrtho(0,self.winfo_width(),self.winfo_height(),0,-1,1) # gluOrtho2d does this, but I'm scared that made my box crash once

    def _redraw(self, *dummy):
            # This must be outside of a pushmatrix, since a resize event
            # will call redraw recursively. 
            self.update_idletasks()
            self.makecurrent()

            self.redraw()
            glFlush()

            self.tk.call(self._w, 'swapbuffers')

    def makecurrent(self):
        """switches opengl context to this widget"""
        self.tk.call(self._w, 'makecurrent')


    def clearscreen(self,color):
        color=(list(color)+[0,0,0,0])[:4] # pad with 0
        glClearColor(*color)
        glClear(GL_COLOR_BUFFER_BIT)

