
""" Many GUIs for wildly different types of data share the same rules
about making selections. Selectable objects ('sob's, from now on) can
be selected and deselected, and operations can be performed on the
selected set. Most GUIs have rules about how to use the mouse or
keyboard to put objects in and out of the selected set. This module
attempts to encode the rules.

Specifically, your new GUI code can feed events into this module (and
the module can ask questions about your GUI objects), and the module
tells you what to put in the selection.

This module is not specific to any GUI toolkit or any larger
program. The module attempts to ask as little as possible of your
program, and you are welcome to disregard any instructions returned by
this module. Your program owns and manages the selection, which is
essentially a set of sobs.

Some systems can use 'continuous' selections, which aren't whole
python objects (or are too many objects). Consider an audio file
editor where we can select ranges of time. (Pretend selecting
individual samples is prohibitive.) Selectionfeel does not yet support
these selections.

"""

from cuisine.utils import make_attributes_from_args

class Selectionfeel:

    def __init__(self,
                 selectionempty,
                 clearselection,
                 addtoselection,
                 ):
        
        """provide selectionfeel with the methods it will need to test
        the selection.

        Examples are shown for the case that your selection is really
        a list called 'sel' (a case so common, it should be acceptable
        in an alternate __init__ form).

        Your functions should probably update the view as well as
        merely adjust the selection list.

        selectionempty - returns true if the selection is empty (len(sel)==0)
        clearselection - makes no objs in the selection (sel[:]=[])
        addtoselection - add given sob to sel (sel.append(sob))
        """
        make_attributes_from_args('selectionempty','clearselection',
                                  )

    def click_sob(self,sob):
        
        """call this if the user clicks (B1 press-release, no
        modifiers) on a sob"""
        
        if not self.selectionempty():
            self.clearselection()
        self.addtoselection(sob)


    def click_nothing(self):
        
        """call this if the user clicks (B1 press-release, no mods)
        outside any sobs

        obviously, you'd only call this method if you wish to have
        this click be considered as adjusting the selection. if the
        click has some other meaning for your program, and it
        shouldn't affect the selection, don't call this method at all.
        """
        self.clearselection()

    def shift_click_sob(self,sob):
        pass
    def ctrl_click_sob(self,sob):
        pass
    def start_drag_sob(self,sob):
        pass
    def start_drag_nothing(self):
        pass
    def motion(self,position):
        #position is passed right back to sob_at_position
        pass
