from __future__ import division
from cuisine.utils import make_attributes_from_args
from cuisine.toss2.timeline.nodeview import Nodeview
import Tkinter as tk
import cuisine.toss2.ogldraw as draw
from cuisine.desctypes import Duration
from dispatch import dispatcher

import logging
log=logging.getLogger("brickview")

from cuisine.player import Decoder
decoder=Decoder()

totalimgs=0

class Brickview(Nodeview):
    """

    one track in the timeline. this is mostly a tk widget on the track
    guide at the left side of the screen, but a trackview also
    involves some opengl material.
    """

    def __init__(self,tkmaster,openglmaster,layout,brick):
        """

        this view is tied to the given track object. the tkmaster is
        the left-side tracks pane.  the given Layout object will tell
        us where the track panel is located and how big it is. 
        
        
        """
        Nodeview.__init__(self,tkmaster,openglmaster,layout,brick)
        self.brick=self.node


        self.highlight=0 # are we lit up
        self.makeimages()

        self.redraw()

    def makeimages(self):
        """builds self.imgs for this clip"""

        global totalimgs
        b=self.brick

        # disable readahead temporarily
        from cuisine import framepool
        oldreadahead=framepool.readaheadframes[0]
        framepool.readaheadframes[0]=0
        
        self.imgs=[]
        try:
            for d in [Duration(fr,'frames') for fr in range(b.intime().getas('frames'),
                                                            b.outtime().getas('frames'),30*5)]:
                if totalimgs>100:
                    continue
                vid=b.tbd_getframe(d,'video')
                totalimgs+=1
                if vid:
                    decoder.decode(vid)
                    self.imgs.append((d,draw.image(rgbstring=decoder.rgbout(),size=(720,480))))
                    log.info("read image %s %s" % ((b.source() and b.source().filename())[:-20],d))
                else:
                    raise "no frame"
        except:
            log.info("image read failed for %r" % b)

        framepool.readaheadframes[0]=oldreadahead

    def drawimages(self):
        """draws thumbnail images onto the brick"""
        layout=self.layout
        track=self.brick.track()
        origin=self.brick.origin()
        y1,y2=layout.windowy_for_track(track)
        scl=.8*(y2-y1)/480
        lastx2=-1000
        if scl>.05:
            for t,i in self.imgs:
                x=layout.windowx_for_showtime(origin+t)
                if x>lastx2+10:
                    i.draw((x,y2-(y2-y1)*.15),scale=scl)
                    lastx2=x+640*scl
        

    def redraw(self):
        Nodeview.redraw(self)

        brick=self.brick
        layout=self.layout

        origin,intime,outtime,track=(brick.origin(),brick.intime(),
                                     brick.outtime(),brick.track())

        x1=layout.windowx_for_showtime(origin+intime)
        x2=layout.windowx_for_showtime(origin+outtime)
        if x2>0 and x1<self.openglmaster.winfo_width():
            y1,y2=layout.windowy_for_track(track)
            if self.highlight:
                bgcolor=(1,.6,.6)
            else:
                bgcolor=(.7,.7,.7)
            draw.rect((x1,y1+1),
                      (x2,y2-1),
                      fillcolor=bgcolor,outlinecolor=(0,0,0))

            self.drawimages()

            draw.text((layout.windowx_for_showtime(origin+intime)+20,y1+35),
                      "%s: %s - %s" % (brick.source() and
                                       brick.source().filename()[-20:],
                                       brick.intime(),brick.outtime()))
    def enter(self,ev):
        """mouse entered this brick"""

        self.highlight=1
        dispatcher.send("brick changed",brickview=self)

    def leave(self,ev):
        """mouse left this brick"""
        self.highlight=0
        dispatcher.send("brick changed",brickview=self)
