from __future__ import generators,nested_scopes,division
from cuisine.toss2.timeline.nodeview import Nodeview
from cuisine.toss2.trackview import Trackview
from brickview import Brickview
from cuisine.utils import make_attributes_from_args

import new
from dispatch import dispatcher
import Tkinter as tk
from cuisine.toss2.ogldisplay import opengldisplay

from skim.zooming import Pair

# from X11
_event_types={4:'ButtonPress',5:'ButtonRelease',6:'Motion'}

class Showview(Nodeview):
    
    """this is the view of an entire show, which contains trackviews
    and brickviews.

    showview also creates the opengl widget and it watches the mouse
    movements over that widget. mouse events are sent to the brickview
    objects via ordinary method calls."""

    show=None # the Show we follow
    trackviews=None # Track : Trackview
    brickviews=None # Brick : Brickview

    def __init__(self,tkmaster,layout,show,commands):
        Nodeview.__init__(self,tkmaster,None,layout,show)
        self.show=self.node
        make_attributes_from_args('commands')

        self.tracknames=tk.Frame(tkmaster,bd=1,relief='raised',width=150)
        self.tracknames.pack(side="left",fill='y')

        # this seems to stop the opengl from being the wrong x visual (palette)
        tkmaster.wait_visibility()
        
        td=self.timelinedisplay=opengldisplay(tkmaster,width=200,height=200,
                                              double=1)
        td.pack(side='right',fill='both',expand=1)

        # opengldisplay needs a 'redraw' method installed
        showview_self=self
        def redraw(self):
            # this becomes a method of an opengldisplay
            self.clearscreen((.8,.8,.8)) # should get bgcolor from show
            showview_self.redraw()
        td.redraw=new.instancemethod(redraw,td,td.__class__)

        for evname in ('<Motion>','<ButtonPress-1>',
                       '<ButtonPress-3>','<ButtonRelease>'):
            td.bind(evname,self.mouseevent)

        if 0: # time scale along the bottom
            timescale=tk.Frame(tkmaster,bd=1,relief='groove')
            timescale.pack(side='bottom')
            for n in range(60):
                tk.Label(timescale,text='|').pack(side='left')

        self.trackviews={}
        for tr in show.tracks().getall():
            self.trackviews[tr]=Trackview(self.tracknames,td,layout,tr)

        self.brickviews={} # Brick : Brickview
        for br in show.bricks().getall():
            self.brickviews[br]=Brickview(self.tracknames,td,layout,br)

        self.setuppanzoom()



        def mkey(ev):
            l=tk.Label(root,text='hello world',width=20,height=3)
            l.place(x=300,y=100)
        #timelinedisplay.bind("<Button-1>",mkey)
    

    def setuppanzoom(self):

        """binds the MMB and C-MMB to the layout methods so you can pan
        and zoom with the mouse"""

        td=self.timelinedisplay
        layout=self.layout
        
        td.bind("<ButtonPress-2>",lambda ev: layout.adjustoffset_start(ev.x))
        td.bind("<B2-Motion>",lambda ev: layout.adjustoffset_move(ev.x))

        td.bind("<Control-ButtonPress-2>",lambda ev: layout.adjustzoom_start(ev.x))
        td.bind("<Control-B2-Motion>",lambda ev: layout.adjustzoom_move(ev.x))
    
    def redraw(self):

        """through opengldisplay trickiness, spelled out in __init__,
        this is called by opengldisplay's _redraw method. so, the
        opengl context is set correctly when this method gets run"""
        
        Nodeview.redraw(self)

        for bv in self.brickviews.values():
            bv.redraw()

        for tv in self.trackviews.values():# (same as layout.alltrackviews())
            tv.redraw()

    def mouseevent(self,ev):

        """called for all mouse events, finds times/brick/track and
        calls the right method for the particular mouse event"""
        
        layout=self.layout
        
        track=layout.track_for_windowy(ev.y)
        t=layout.showtime_for_windowx(ev.x)
        
        brick=self.show.bricks().gettrackbrick(track,t)

        # decorate the event a little so it's easier to read in the
        # later methods
        ev.b1down=ev.state&256
        ev.b3down=ev.state&1024

        # call a method named after the event name
        if int(ev.type) in _event_types:
            getattr(self,_event_types[int(ev.type)])(ev,track,t,brick)
            
    def Motion(self,ev,track,t,brick):
        if not ev.b1down and not ev.b3down:
            # enter/leave objects
            enterobj=None
            if brick:
                enterobj=self.brickviews[brick]
            lastenter=getattr(self,"_lastenter",None)
            if enterobj!=lastenter:
                if lastenter is not None: self._lastenter.leave(ev)
                if enterobj is not None: enterobj.enter(ev)
                self._lastenter=enterobj

        if ev.b1down:
            print "dragging selection"
            
        
    def ButtonPress(self,ev,track,t,brick):
        print "clicked %s at %s" % (brick,t)
        self._dragstart=Pair(ev.x,ev.y)
        self._timestart=t
        
    def ButtonRelease(self,ev,track,t,brick):
        print "release %s at %s" % (brick,t)
        dragend=Pair(ev.x,ev.y)
        if dragend.dist(self._dragstart)<5: # 'closeenough'
            # clicked in-place
            if brick:
                self.commands.newcommand("addToSelection(%s)"%brick)
            else:
                self.commands.newcommand("clearSelection()")

        if brick:
            brick.origin(brick.origin()+t-self._timestart)
            dispatcher.send("brick changed",brickview=self.brickviews[brick])
        
