"""extra tools for all programs. this is for import *"""

from __future__ import generators, division
import sys,os,time,commands,logging,xml


def localfile(filename):
    the_dir_of_this_module=os.path.dirname(__file__)
    return os.path.join(the_dir_of_this_module,filename)
    
def createuuid():
    """returns a universally unique id as a string. string is limited
    to hex and dashes"""
    u=commands.getoutput("uuidgen")
    u.strip()
    return u
    
def commonlogsetup(excludechannels=[]):
    """sets up logs how i usually like them, with a custom filter to
    exclude the given channel names (which means set them to level
    INFO in spite of the main log level setting). log level is DEBUG,
    but you can change that later on your own. return value is the
    top-level logger"""
    import logging
    if isinstance(excludechannels,(str,unicode)):
        excludechannels=[excludechannels]
    log=logging.getLogger()
    log.setLevel(logging.DEBUG)
    h=logging.StreamHandler()
    h.setFormatter(logging.Formatter("%(asctime)s %(levelname)-5s %(name)s %(filename)s:%(lineno)d: %(message)s"))

    if 1:
        for chan in excludechannels:
            logging.getLogger(chan).setLevel(logging.INFO)
    else:
        class myfilter(logging.Filter):
            def filter(self,record,excludechannels=excludechannels):
                return record.name not in excludechannels
                # return (record.levelno>logging.DEBUG) or (record.name not in excludechannels)
        h.addFilter(myfilter())

    log.addHandler(h)
    return log

def runstats(cmd="go()",statfile="profile.stats"):
    """
    the new version should be like this:
    
        import hotshot, hotshot.stats, os
        from tempfile import mktemp
        pfile=mktemp("profile")
        prof=hotshot.Profile(pfile,lineevents=1)
        prof.runcall(main,options,args)
        prof.close()
        stats = hotshot.stats.load(pfile)
        stats.strip_dirs()
        stats.sort_stats('time', 'calls')
        tats.print_stats(20)
        os.unlink(pfile)


    """
    import profile
    profile.run(cmd,statfile)
    import pstats
    p=pstats.Stats(statfile)
    #p.strip_dirs()
    
    p.sort_stats('time')#'cumulative') #0 or 'time'
    p.print_stats(60)
    p.print_callers(40)


def enumerate(*collections):
    """Generates an indexed series:  (0,coll[0]), (1,coll[1]) ...
    
    this is a multi-list version of the code from the PEP:
    enumerate(a,b) gives (0,a[0],b[0]), (1,a[1],b[1]) ...
    """
    i = 0
    its = [iter(collection) for collection in collections]
    while 1:
        yield [i,]+[it.next() for it in its]
        i += 1
        
def make_attributes_from_args(*argnames):
    """
    This function simulates the effect of running
      self.foo=foo
    for each of the given argument names ('foo' in the example just
    now). Now you can write:
        def __init__(self,foo,bar,baz):
            make_attributes_from_args('foo','bar','baz')
            ...
    instead of:
        def __init__(self,foo,bar,baz):
            self.foo=foo
            self.bar=bar
            self.baz=baz
            ... 
    """
    
    callerlocals=sys._getframe(1).f_locals
    callerself=callerlocals['self']
    for a in argnames:
        setattr(callerself,a,callerlocals[a])
        
    
    """
    ----------------------------------------------------------------
    More discussion:
    
    I feel that the non-standard function (this one) is justified even
    though it is less clear than the comb code above. The comb code
    version is as offensive to me as this is:
      mylist.append('a')
      mylist.append('b')
      mylist.append('c')
      
    That code, while clear, punishes both the reader and programmer.
    I hope that none of you would ever write the above, and that you'd
    use something like "mylist.extend(['a','b','c'])" instead. My
    make_attributes_from_args function is to be used in an analagous
    manner. (Yes, certain situations will warrant code that is
    repetitive in some way or another. But I'm talking about
    self.foo=foo stuff, not those situations.)

    Another way to look at the attribute-assignment problem: There's
    only one 'program step' going on in the block-- some locals are
    getting copied to instance variables. One thing deserves one line.

    The disadvantage of the function is that it makes ordinary code
    harder to read for programmers who haven't read these
    instructions.  My only weapon against that problem is the choice
    of the function's name.  I deliberately chose 'make' and 'args'
    instead of 'set' and 'locals' because I want the reader to
    immediately think of the new attributes that are (probably) being
    created out of the arguments to __init__. The fact that this
    function will operate on any locals and set existing attributes to
    new values is incidental.

    I would be interested to hear better names for the function.  If I
    get a better one, I will convert all my code immediately.

    For another look at the same problem, see:
      http://twistedmatrix.com/users/acapnotic/keeparg.py.html
    """


class ClassyDict(dict):
    """a dict that accepts attribute-style access as well (for keys
    that are legal names, obviously). I used to call this Struct, but
    chose the more colorful name to avoid confusion with the struct
    module."""
    def __getattr__(self,a):
        return self[a]
    def __setattr__(self,a,v):
        self[a]=v
    def __delattr__(self,a):
        del self[a]
        

def dumpobj(o):
    """prints all the object's non-callable attributes"""
    print repr(o)
    for a in [x for x in dir(o) if not callable(getattr(o,x))]:
        try:
            print "  %20s: %s " % (a,getattr(o,a))
        except:
            pass
    print ""
 


# from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/105873
# I added the symlink option
def dirwalk(dir,symlinkok=0):
    "walk a directory tree, using a generator"
    for f in os.listdir(dir):
        fullpath = os.path.join(dir,f)
        if os.path.isdir(fullpath) and (symlinkok or not os.path.islink(fullpath)):
            for x in dirwalk(fullpath,symlinkok):  # recurse into subdir
                yield x
        else:
            yield fullpath

class Condition:

    """mutable boolean which outputs log messages when it
    changes. please suggest better names for this class, possibly
    including 'hysteresis', 'transition', or 'edge'. pass it your
    logger object."""
    
    def __init__(self,truemsg="condition became true",
                 falsemsg="condition became false",log=None,level=logging.INFO):
        self.lastval=None
        self.falsemsg=falsemsg
        self.truemsg=truemsg
        self.level=level
        self.log=log
    def set(self,newval,*args):
        """changes the internal value, and if there's a transition
        between true and false, the appropriate log message is
        output. if your log message has formatters in it, pass their
        values to the set() call"""
        if self.log:
            if newval>self.lastval:
                self.log.log(self.level,self.truemsg,*args)
            if newval<self.lastval:
                self.log.log(self.level,self.falsemsg,*args)
        self.lastval=newval
    def setquiet(self,newval):
        """changes the value with no logging"""
        self.lastval=newval
    def clear(self):
        self.setquiet(0)

def stable_unique(l):

    """returns a copy of l in order, but without any elements that are
    equal to earlier elements"""

    u=[]
    seen={}
    for obj in l:
        if obj not in seen:
            seen[obj]=obj
            u.append(obj)
    return u

def isxml(fileobj):
    
    """returns whether the file appears to be xml by having SAX read a
    little bit of it and report any errors. the point of this test,
    rather than feeding the file to DescFileManager, is that this
    function aborts fast on a giant file. DFM would read the whole
    thing."""
    import xml.sax

    parser = xml.sax.make_parser()
    parser._bufsize=1024
    try:
        parser.parse(fileobj)
        return 1
    except xml.sax._exceptions.SAXParseException:
        return 0
    
